'''GP regression in 1D, with noisy evaluations

This example repeats example02 with a noisy dataset.

----
Author: Emmanuel Vazquez <emmanuel.vazquez@centralesupelec.fr>
Copyright (c) 2022-2024, CentraleSupelec
License: GPLv3 (see LICENSE)
----
This example is based on the file stk_example_kb01.m from the STK at
https://github.com/stk-kriging/stk/ 
by Julien Bect and Emmanuel Vazquez, released under the GPLv3 license.

Original copyright notice:

   Copyright (c) 2015, 2016, 2018 CentraleSupelec
   Copyright (c) 2011-2014 SUPELEC
----

'''
import math
import numpy as np
import gpmp.num as gnp
import gpmp as gp

def generate_data(noise_std):
    """
    Generates a noisy dataset

    Args:
        noise_std (float): Standard deviation of noise.

    Returns:
        tuple: (xt, zt, xi, zi) - target and input datasets.
    """
    dim = 1
    nt = 200
    box = [[-1], [1]]
    xt = gp.misc.designs.regulargrid(dim, nt, box)
    zt = gp.misc.testfunctions.twobumps(xt)

    ind = [10, 45, 100, 130, 130, 130, 131, 132, 133, 133, 133, 134, 160]
    xi = xt[ind]
    zi = zt[ind] + noise_std * np.random.randn(len(ind))

    return xt, zt, xi, zi


def constant_mean(x, _):
    return gnp.ones((x.shape[0], 1))


def kernel_ii_or_tt(x, param, pairwise=False):
    p = 2
    sigma2 = gnp.exp(param[0])
    loginvrho = param[1]
    noise_variance = gnp.exp(param[2])

    if pairwise:
        K = sigma2 * gnp.ones((x.shape[0], ))
    else:
        K = gnp.scaled_distance(loginvrho, x, x)
        K = sigma2 * gp.kernel.maternp_kernel(p, K) + noise_variance * gnp.eye(K.shape[0])

    return K


def kernel_it(x, y, param, pairwise=False):
    p = 2
    sigma2 = gnp.exp(param[0])
    loginvrho = param[1]

    if pairwise:
        K = gnp.scaled_distance_elementwise(loginvrho, x, y)
    else:
        K = gnp.scaled_distance(loginvrho, x, y)

    K = sigma2 * gp.kernel.maternp_kernel(p, K)
    return K


def kernel(x, y, param, pairwise=False):
    if y is x or y is None:
        return kernel_ii_or_tt(x, param, pairwise)
    else:
        return kernel_it(x, y, param, pairwise)


def main():
    noise_std = 1e-1
    xt, zt, xi, zi = generate_data(noise_std)

    mean = constant_mean
    meanparam = None

    covparam = gnp.array([
        math.log(0.5**2),
        math.log(1 / .7),
        2 * math.log(noise_std)])

    model = gp.core.Model(mean, kernel, meanparam, covparam)

    (zpm, zpv) = model.predict(xi, zi, xt)

    return xt, zt, xi, zi, zpm, zpv


def visualize(xt, zt, xi, zi, zpm, zpv):
    fig = gp.misc.plotutils.Figure(isinteractive=True)
    fig.plot(xt, zt, 'C0', linestyle=(0, (5, 5)), linewidth=1)
    fig.plot(xi, zi, 'rs')
    fig.plotgp(xt, zpm, zpv)
    fig.xylabels('x', 'z')
    fig.title('GP regression with noisy observations')
    fig.show(grid=True)


if __name__ == "__main__":
    xt, zt, xi, zi, zpm, zpv = main()
    visualize(xt, zt, xi, zi, zpm, zpv)



"""
Author: Emmanuel Vazquez <emmanuel.vazquez@centralesupelec.fr>
Copyright (c) 2022-2024, CentraleSupelec
License: GPLv3 (see LICENSE)
--------------------------------------------------------------

Description:
    This module sets and prints the backend to be used by the GPmp (Gaussian
    Process micro package) framework.

    The backend is set by checking for the existence of certain libraries in
    the system and, if available, sets an environment variable "GPMP_BACKEND" 
    to the corresponding library's name.

    The priority order for backends is: torch -> jax -> numpy. If neither 
    torch nor jax are found in the system, numpy is set as the default backend.

"""

import os
import numpy
from importlib import util as importlib_util

_gpmp_backend_ = os.environ.get("GPMP_BACKEND")


def set_backend_env_var(backend):
    global _gpmp_backend_
    os.environ["GPMP_BACKEND"] = backend
    _gpmp_backend_ = backend


# Automatically set the backend if not already set in the environment.
if _gpmp_backend_ is None:
    if importlib_util.find_spec("torch") is not None:
        set_backend_env_var("torch")
    elif importlib_util.find_spec("jax") is not None:
        set_backend_env_var("jax")
    else:
        set_backend_env_var("numpy")

print(f"Using backend: {_gpmp_backend_}")


# -----------------------------------------------------
#
#                      NUMPY
#
# -----------------------------------------------------
if _gpmp_backend_ == "numpy":
    from numpy import array, empty

    from numpy import (
        copy,
        reshape,
        where,
        any,
        isscalar,
        isnan,
        isinf,
        isfinite,
        unique,
        hstack,
        vstack,
        stack,
        tile,
        concatenate,
        expand_dims,
        empty,
        empty_like,
        zeros,
        ones,
        full,
        full_like,
        eye,
        diag,
        arange,
        linspace,
        logspace,
        meshgrid,
        abs,
        sqrt,
        exp,
        log,
        log10,
        sum,
        mean,
        std,
        cov,
        sort,
        min,
        max,
        argmin,
        argmax,
        minimum,
        maximum,
        einsum,
        matmul,
        inner,
        all,
        logical_not,
        logical_and,
        logical_or,
    )
    from numpy.linalg import norm, cond, cholesky, qr, svd, inv
    from numpy.random import rand, randn, choice
    from numpy import pi, inf
    from numpy import finfo, float64
    from scipy.special import gammaln
    from scipy.linalg import solve, solve_triangular, cho_factor, cho_solve
    from scipy.spatial.distance import cdist
    from scipy.stats import norm as normal
    from scipy.stats import multivariate_normal as scipy_mvnormal

    eps = finfo(float64).eps
    fmax = numpy.finfo(numpy.float64).max

    def set_elem1(x, index, v):
        x[index] = v
        return x

    def set_row2(A, index, x):
        A[index, :] = x
        return A

    def set_col2(A, index, x):
        A[:, index] = x
        return A

    def set_col3(A, index, x):
        A[:, :, index] = x
        return A

    def asarray(x):
        if isinstance(x, numpy.ndarray):
            return x
        elif isinstance(x, (int, float)):
            return numpy.array([x])
        else:
            return numpy.asarray(x)

    def asdouble(x):
        return x.astype(float64)

    def to_np(x):
        return x

    def to_scalar(x):
        return x.item()

    def isarray(x):
        return isinstance(x, numpy.ndarray)

    def inftobigf(a, bigf=fmax / 1000.0):
        a = where(numpy.isinf(a), numpy.full_like(a, bigf), a)
        return a

    def grad(f):
        return None

    class jax:
        @staticmethod
        def jit(f, *args, **kwargs):
            return f

    def scaled_distance(loginvrho, x, y):
        invrho = exp(loginvrho)
        xs = invrho * x
        ys = invrho * y
        return cdist(xs, ys)

    def scaled_distance_elementwise(loginvrho, x, y):
        if x is y or y is None:
            d = zeros((x.shape[0],))
        else:
            invrho = exp(loginvrho)
            d = sqrt(sum(invrho * (xs - ys) ** 2, axis=1))
        return d

    def cholesky_inv(A):
        # FIXME: slow!
        # n = A.shape[0]
        # C, lower = cho_factor(A)
        # Ainv = cho_solve((C, lower), eye(n))
        return inv(A)

    def cholesky_solve(A, b):
        L = cholesky(A)
        y = solve_triangular(L, b, lower=True)
        x = solve_triangular(L.T, y, lower=False)
        return x, L

    class multivariate_normal:
        @staticmethod
        def rvs(mean=0.0, cov=1.0, n=1):
            # Check if cov is a scalar or 1x1 array, and use norm if so
            if isscalar(cov) or (isinstance(cov, numpy.ndarray) and cov.size == 1):
                return normal.rvs(mean, numpy.sqrt(cov), size=n)

            # For dxd covariance matrix, use multivariate_normal
            d = cov.shape[0]  # Dimensionality from the covariance matrix
            mean_array = numpy.full(d, mean)  # Expand mean to an array
            return scipy_mvnormal.rvs(mean=mean_array, cov=cov, size=n)

        @staticmethod
        def logpdf(x, mean=0.0, cov=1.0):
            # Check if cov is a scalar or 1x1 array, and use norm if so
            if numpy.isscalar(cov) or (
                isinstance(cov, numpy.ndarray) and cov.size == 1
            ):
                return normal.logpdf(x, mean, numpy.sqrt(cov))

            # For dxd covariance matrix, use multivariate_normal
            d = x.shape[-1] if x.ndim > 1 else 1  # Infer dimensionality from x
            mean_array = numpy.full(d, mean)  # Expand mean to an array
            return scipy_mvnormal.logpdf(x, mean=mean_array, cov=cov)

        @staticmethod
        def cdf(x, mean=0.0, cov=1.0):
            # Check if cov is a scalar or 1x1 array, and use norm for the univariate case
            if isscalar(cov) or (isinstance(cov, numpy.ndarray) and cov.size == 1):
                return normal.cdf(x, mean, sqrt(cov))

            # For dxd covariance matrix, use multivariate_normal
            if isinstance(mean, (float, int)):
                d = cov.shape[0]  # Dimensionality from the covariance matrix
                mean = full(d, mean)  # Expand mean to an array
            return scipy_mvnormal.cdf(x, mean=mean, cov=cov)


# -----------------------------------------------------
#
#                      TORCH
#
# -----------------------------------------------------
elif _gpmp_backend_ == "torch":
    import torch

    torch.set_default_dtype(torch.float64)

    from torch import tensor, is_tensor

    from torch import (
        reshape,
        where,
        any,
        isnan,
        isinf,
        isfinite,
        hstack,
        vstack,
        stack,
        tile,
        concatenate,
        empty,
        empty_like,
        zeros,
        ones,
        full,
        eye,
        diag,
        arange,
        linspace,
        logspace,
        meshgrid,
        abs,
        cov,
        argmax,
        argmin,
        einsum,
        matmul,
        inner,
        logical_not,
        logical_and,
        logical_or,
    )
    from torch.linalg import cond, cholesky, qr, inv
    from torch import rand, randn
    from torch import pi, inf
    from torch import finfo, float64
    from torch.distributions.multivariate_normal import MultivariateNormal
    from torch.distributions.normal import Normal
    from scipy.stats import multivariate_normal as scipy_mvnormal
    from scipy.special import gammaln

    eps = finfo(float64).eps
    fmax = finfo(float64).max

    def copy(x):
        return torch.clone(tensor(x))

    def set_elem1(x, index, v):
        x[index] = v
        return x

    def set_row2(A, index, x):
        A[index, :] = x
        return A

    def set_col2(A, index, x):
        A[:, index] = x
        return A

    def set_col3(A, index, x):
        A[:, :, index] = x
        return A

    def expand_dims(tensor, axis):
        return tensor.unsqueeze(axis)

    def array(x: list):
        return tensor(x)

    def asarray(x):
        if isinstance(x, torch.Tensor):
            return x
        elif isinstance(x, (int, float)):
            return tensor([x])
        else:
            return torch.asarray(x)

    def asdouble(x):
        return x.to(torch.double)

    def to_np(x):
        return x.numpy()

    def to_scalar(x):
        return x.item()

    def isarray(x):
        return torch.is_tensor(x)

    def isscalar(x):
        if torch.is_tensor(x):
            return x.reshape(-1).size()[0] == 1
        elif isinstance(x, (int, float)):
            return True
        else:
            return False

    def scalar_safe(f):
        def f_(x):
            if torch.is_tensor(x):
                return f(x)
            else:
                return f(torch.tensor(x))

        return f_

    log = scalar_safe(torch.log)
    log10 = scalar_safe(torch.log10)
    exp = scalar_safe(torch.exp)
    sqrt = scalar_safe(torch.sqrt)

    def axis_to_dim(f):
        def f_(x, axis=None, **kwargs):
            if axis is None:
                return f(x, **kwargs)
            else:
                return f(x, dim=axis, **kwargs)

        return f_

    all = axis_to_dim(torch.all)
    unique = axis_to_dim(torch.unique)
    sum = axis_to_dim(torch.sum)
    mean = axis_to_dim(torch.mean)
    std = axis_to_dim(torch.std)
    var = axis_to_dim(torch.var)

    def norm(x, axis=None, ord=2):
        return torch.norm(x, dim=axis, p=ord)

    def min(x, axis=0):
        m = torch.min(x, dim=axis)
        return m.values

    def max(x, axis=0):
        m = torch.max(x, dim=axis)
        return m.values

    def maximum(x1, x2):
        if not torch.is_tensor(x1):
            x1 = tensor(x1)
        if not torch.is_tensor(x2):
            x2 = tensor(x2)
        return torch.maximum(x1, x2)

    def minimum(x1, x2):
        if not torch.is_tensor(x1):
            x1 = tensor(x1)
        if not torch.is_tensor(x2):
            x2 = tensor(x2)
        return torch.minimum(x1, x2)

    def sort(x, axis=-1):
        xsorted = torch.sort(x, dim=axis)
        return xsorted.values

    def inftobigf(a, bigf=fmax / 1000.0):
        a = torch.where(torch.isinf(a), torch.full_like(a, bigf), a)
        return a

    def svd(A, full_matrices=True, hermitian=True):
        return torch.linalg.svd(A, full_matrices)

    def solve(A, B, overwrite_a=True, overwrite_b=True, assume_a="gen", sym_pos=False):
        return torch.linalg.solve(A, B)

    def grad(f):
        def f_grad(x):
            if not torch.is_tensor(x):
                x = torch.tensor(x, requires_grad=True)
            else:
                x = x.detach().clone().requires_grad_(True)

            y = f(x)
            gradients = torch.autograd.grad(y, x, allow_unused=True)[0]
            return gradients

        return f_grad

    class jax:
        @staticmethod
        def jit(f, *args, **kwargs):
            return f

    def custom_sqrt(x):
        arbitrary_value = 1.0
        mask = x == 0.0
        x_copy = torch.where(mask, arbitrary_value, x)

        res = torch.where(mask, 0.0, sqrt(x_copy))

        return res

    def cdist(x, y, zero_diagonal=True):
        if x is y:
            # use view method: requires contiguous tensor
            x_norm = (x**2).sum(1).view(-1, 1)
            distances = x_norm + x_norm.t() - 2.0 * torch.mm(x, x.t())
        else:
            x_norm = (x**2).sum(1).view(-1, 1)
            y_norm = (y**2).sum(1).view(1, -1)
            distances = x_norm + y_norm - 2.0 * torch.mm(x, y.t())

        distances = custom_sqrt(distances.clamp(min=0.0))

        if zero_diagonal and x is y:
            mask = torch.eye(distances.size(0), dtype=torch.bool, device=x.device)
            distances = distances.masked_fill(mask, 0.0)

        return distances

    def scaled_distance(loginvrho, x, y):
        invrho = exp(loginvrho)
        xs = invrho * x

        if x is y:
            d = cdist(xs, xs)
        else:
            ys = invrho * y
            d = cdist(xs, ys)

        return d

    def scaled_distance_elementwise(loginvrho, x, y):
        if x is y or y is None:
            d = zeros((x.shape[0],))
        else:
            invrho = exp(loginvrho)
            d = sqrt(sum(invrho * (xs - ys) ** 2, axis=1))
        return d

    def cho_factor(A, lower=False, overwrite_a=False, check_finite=True):
        # torch.linalg does not have cho_factor(), use cholesky() instead.
        return cholesky(A, upper=not (lower))

    def cholesky_solve(A, b):
        if b.dim() == 1:
            b = b.reshape(-1, 1)
        L = cholesky(A)
        y = torch.linalg.solve_triangular(L, b, upper=False)
        x = torch.linalg.solve_triangular(L.t(), y, upper=True)
        return x, L

    def cholesky_inv(A):
        C = cholesky(A)
        return torch.cholesky_inverse(C)

    class normal:
        @staticmethod
        def cdf(x, loc=0.0, scale=1.0):
            d = Normal(loc, scale)
            return d.cdf(x)

        @staticmethod
        def logcdf(x, loc=0.0, scale=1.0):
            d = Normal(loc, scale)
            return log(d.cdf(x))

        @staticmethod
        def pdf(x, loc=0.0, scale=1.0):
            t = (x - loc) / scale
            return 1 / sqrt(2 * pi) * exp(-0.5 * t**2)

        @staticmethod
        def logpdf(x, loc=0.0, scale=1.0):
            d = Normal(loc, scale)
            return d.logpdf(x)

    class multivariate_normal:
        @staticmethod
        def rvs(mean=0.0, cov=1.0, n=1):
            # Check if cov is a scalar or 1x1 array, and use Normal if so
            if (
                torch.is_tensor(cov)
                and cov.ndim == 0
                or (cov.ndim == 2 and cov.shape[0] == 1 and cov.shape[1] == 1)
            ):
                distribution = Normal(torch.tensor(mean), cov.sqrt())
                return distribution.sample((n,))

            # For dxd covariance matrix, use MultivariateNormal
            d = cov.shape[0]  # Dimensionality from the covariance matrix
            mean_tensor = torch.full((d,), mean)  # Expand mean to a tensor
            distribution = MultivariateNormal(mean_tensor, covariance_matrix=cov)
            return distribution.sample((n,))

        @staticmethod
        def logpdf(x, mean=0.0, cov=1.0):
            # Check if cov is a scalar or 1x1 array, and use Normal if so
            if (
                torch.is_tensor(cov)
                and cov.ndim == 0
                or (cov.ndim == 2 and cov.shape[0] == 1 and cov.shape[1] == 1)
            ):
                distribution = Normal(torch.tensor(mean), cov.sqrt())
                return distribution.log_prob(x.squeeze(-1))

            # For dxd covariance matrix, use MultivariateNormal
            d = x.shape[-1]  # Infer dimensionality from x
            mean_tensor = torch.full((d,), mean)  # Expand mean to a tensor
            distribution = MultivariateNormal(mean_tensor, covariance_matrix=cov)
            return distribution.log_prob(x)

        @staticmethod
        def cdf(x, mean=0.0, cov=1.0):
            # Convert inputs to NumPy arrays if they are PyTorch tensors
            if torch.is_tensor(x):
                x = x.numpy()
            if torch.is_tensor(mean):
                mean = mean.numpy()
            if torch.is_tensor(cov):
                cov = cov.numpy()

            # Check if cov is a scalar or 1x1 array, and use norm for the univariate case
            if (
                isscalar(cov)
                or (isinstance(cov, numpy.ndarray) and cov.size == 1)
                or (torch.is_tensor(cov) and cov.size == 1)
            ):
                return Normal.cdf(x, mean, np.sqrt(cov))

            # For dxd covariance matrix, use multivariate_normal
            if isscalar(mean):
                d = cov.shape[0]  # Dimensionality from the covariance matrix
                mean = full(d, mean)  # Expand mean to an array
            return scipy_mvnormal.cdf(x, mean, cov)


# ------------------------------------------------------
#
#                        JAX
#
# ------------------------------------------------------
elif _gpmp_backend_ == "jax":
    import jax

    # set multithreaded/multicore parallelism
    # see https://github.com/google/jax/issues/8345
    # os.environ["XLA_FLAGS"] = "--xla_force_host_platform_device_count=512"

    # set double precision for floats
    jax.config.update("jax_enable_x64", True)

    from jax.numpy import array, empty

    from jax.numpy import (
        reshape,
        where,
        any,
        isscalar,
        isnan,
        isinf,
        isfinite,
        unique,
        hstack,
        vstack,
        stack,
        tile,
        concatenate,
        expand_dims,
        empty,
        empty_like,
        zeros,
        ones,
        full,
        full_like,
        eye,
        diag,
        arange,
        linspace,
        logspace,
        meshgrid,
        abs,
        sqrt,
        exp,
        log,
        log10,
        sum,
        mean,
        std,
        cov,
        sort,
        min,
        max,
        argmin,
        argmax,
        minimum,
        maximum,
        einsum,
        matmul,
        inner,
        all,
        logical_not,
        logical_and,
        logical_or,
    )
    from jax.numpy.linalg import norm
    from jax.numpy.linalg import cond, cholesky, qr, svd, inv
    from jax.numpy import pi, inf
    from jax.numpy import finfo, float64
    from jax.scipy.special import gammaln
    from jax.scipy.linalg import solve, solve_triangular, cho_factor, cho_solve
    from jax.scipy.stats import norm as normal
    from scipy.stats import multivariate_normal as scipy_mvnormal

    eps = finfo(float64).eps
    fmax = finfo(float64).max

    def copy(x):
        return array(x, copy=True)

    @jax.jit
    def set_elem1(x, index, v):
        return x.at[index].set(v)

    @jax.jit
    def set_row2(A, index, x):
        return A.at[index, :].set(x)

    @jax.jit
    def set_col2(A, index, x):
        return A.at[:, index].set(x)

    @jax.jit
    def set_col3(A, index, x):
        return A.at[:, :, index].set(x)

    def asarray(x):
        if isinstance(x, jax.numpy.ndarray):
            return x
        if isinstance(x, (int, float)):
            return jax.numpy.array([x])
        else:
            return jax.numpy.asarray(x)

    def asdouble(x):
        return x.astype(float64)

    def to_np(x):
        return numpy.array(x)

    def to_scalar(x):
        return x.item()

    def isarray(x):
        return isinstance(x, jax.numpy.ndarray)

    def inftobigf(a, bigf=fmax / 1000.0):
        a = where(isinf(a), full_like(a, bigf), a)
        return a

    from jax import grad

    def cdist(x, y):
        if y is None:
            y = x
        y2 = sum(y**2, axis=1)
        # Debug: check if x is y
        # print("&x = {}, &y = {}".format(hex(id(x)), hex(id(y))))
        if x is y:
            d = sqrt(reshape(y2, [-1, 1]) + y2 - 2 * inner(x, y))
        else:
            x2 = reshape(sum(x**2, axis=1), [-1, 1])
            d = sqrt(x2 + y2 - 2 * inner(x, y))
        return d

    @jax.custom_vjp
    def scaled_distance_(loginvrho, x, y):
        invrho = exp(loginvrho)
        xs = invrho * x
        ys = invrho * y
        hs2 = (xs - ys) ** 2
        d = sqrt(sum(hs2))
        return d

    def scaled_distance__fwd(loginvrho, x, y):
        invrho = exp(loginvrho)
        xs = invrho * x
        ys = invrho * y
        hs2 = (xs - ys) ** 2
        d = sqrt(sum(hs2))
        intermediates = (d, hs2)
        return d, intermediates

    def scaled_distance__bwd(intermediates, g):
        d, hs2 = intermediates
        grad_loginvrho = g * hs2 / (d + eps)
        grad_x = None  # Not computed, since we only need the gradient with respect to loginvrho
        grad_y = None
        return (grad_loginvrho, grad_x, grad_y)

    scaled_distance_.defvjp(scaled_distance__fwd, scaled_distance__bwd)

    def scaled_distance(loginvrho, x, y):
        f = jax.vmap(
            lambda x1: jax.vmap(
                lambda y1: jax.jit(scaled_distance_)(loginvrho, x1, y1)
            )(y)
        )(x)
        return f

    def scaled_distance_elementwise(loginvrho, x, y):
        f = jax.vmap(jax.jit(scaled_distance), in_axes=(None, 0, 0), out_axes=0)(
            loginvrho, x, y
        )
        return f

    def cholesky_inv(A):
        # FIXME: slow!
        # n = A.shape[0]
        # C, lower = cho_factor(A)
        # Ainv = cho_solve((C, lower), eye(n))
        return inv(A)

    def cholesky_solve(A, b):
        L = cholesky(A)
        y = solve_triangular(L, b, lower=True)
        x = solve_triangular(L.T, y, lower=False)
        return x, L

    def seed(s):
        return jax.random.PRNGKey(s)

    key = seed(0)

    def rand(*args):
        return jax.random.uniform(key, shape=args)

    def randn(*args):
        return jax.random.normal(key, shape=args)

    class multivariate_normal:
        @staticmethod
        def rvs(mean=0.0, cov=1.0, n=1):
            if isscalar(cov) or cov.ndim == 0:
                cov = cov * eye(1)
            if isscalar(mean):
                mean = array([mean])

            d = cov.shape[0]  # Dimensionality from the covariance matrix
            mean_vector = full((d,), mean)  # Expand mean to a vector

            return jax.random.multivariate_normal(
                key, mean=mean_vector, cov=cov, shape=(n,)
            )

        @staticmethod
        def logpdf(x, mean=0.0, cov=1.0, allow_singular=False):
            mean_vector = full(
                (x.shape[-1],), mean
            )  # Expand mean to match x dimensions
            return jax.scipy.stats.multivariate_normal.logpdf(
                x, mean=mean_vector, cov=cov, allow_singular=allow_singular
            )

        @staticmethod
        def cdf(x, mean=0.0, cov=1.0):
            # Convert JAX arrays to NumPy arrays for compatibility with SciPy
            if isinstance(x, jax.numpy.ndarray):
                x = numpy.array(x)
            if isinstance(mean, jax.numpy.ndarray):
                mean = numpy.array(mean)
            if isinstance(cov, jax.numpy.ndarray):
                cov = numpy.array(cov)

            # Check if cov is a scalar or 1x1 array, and use norm for the univariate case
            if isscalar(cov) or (isinstance(cov, jax.numpy.ndarray) and cov.size == 1):
                return normal.cdf(x, mean, sqrt(cov))

            # For dxd covariance matrix, use multivariate_normal
            if isscalar(mean):
                d = cov.shape[0]  # Dimensionality from the covariance matrix
                mean = full(d, mean)  # Expand mean to an array

            return scipy_mvnormal.cdf(x, mean, cov)


# ------------------------------------------------------------------

else:
    raise RuntimeError(
        "Please set the GPMP_BACKEND environment variable to 'numpy', 'torch' or 'jax'."
    )



## --------------------------------------------------------------
# Author: Emmanuel Vazquez <emmanuel.vazquez@centralesupelec.fr>
# Copyright (c) 2022-2023, CentraleSupelec
# License: GPLv3 (see LICENSE)
## --------------------------------------------------------------
import numpy as np
from scipy.stats import qmc
from scipy.spatial.distance import cdist, pdist


def maxdist(sample):
    """
    Calculate the maximum distance (diameter) between any pair of points in the sample.

    Parameters
    ----------
    sample : numpy.ndarray
        Array of points in the sample.

    Returns
    -------
    float
        Maximum distance between any pair of points in the sample.
    """
    D = pdist(sample)
    maxdist = np.max(D)
    return maxdist


def mindist(sample):
    """
    Calculate the minimum distance (separation) between any pair of points in the sample.

    Parameters
    ----------
    sample : numpy.ndarray
        Array of points in the sample.

    Returns
    -------
    float
        Minimum distance between any pair of points in the sample.
    """
    D = pdist(sample)
    mindist = np.min(D)
    return mindist


def discrepancy(sample):
    """
    Calculate the discrepancy of the sample.

    Parameters
    ----------
    sample : numpy.ndarray
        Array of points in the sample.

    Returns
    -------
    float
        Discrepancy value of the sample.
    """
    return qmc.discrepancy(sample)


def filldist_approx(sample, box, n=int(1e6), x=None):
    """
    Approximate the fill distance using a random uniform discretization of the box.

    Parameters
    ----------
    sample : numpy.ndarray
        Array of points in the sample.
    box : list of lists
        List of lists containing the lower and upper bounds of the box.
    n : int, optional
        Number of points in the random uniform discretization, default is 1e6.
    x : numpy.ndarray, optional
        Points in the random uniform discretization, default is None.

    Returns
    -------
    float
        Approximated fill distance.
    """
    dim = sample.shape[1]
    if x is None:
        x = randunif(dim, n, box)
    else:
        n = x.shape[0]
    filldist = 0
    for i in range(n):
        D = cdist(sample, x)
        d = np.min(D)
        if d > filldist:
            filldist = d
    return filldist


def scale(sample_standard, box):
    """
    Map a standard sample in [0, 1]^dim to the given box.

    Parameters
    ----------
    sample_standard : numpy.ndarray
        Array of points in the standard sample.
    box : list of lists
        List of lists containing the lower and upper bounds of the box.

    Returns
    -------
    numpy.ndarray
        Sample points mapped to the given box.
    """
    l_bounds, u_bounds = box[0], box[1]
    sample_box = qmc.scale(sample_standard, l_bounds, u_bounds)
    return sample_box


def regulargrid(dim, n, box):
    """
    Build a regular grid in the dim-dimensional hyperrectangle.

    If n is an integer, a grid of size n^dim is built;

    If n is a list of length dim, a grid of size prod(n) is built,
    with n_i points on coordinate i.

    The dim-dimensional hyperrectangle is specified by the argument
    box, which is a 2 x dim array where box_(1, i) and box_(2, i) are
    the lower- and upper-bound of the interval on the i^th coordinate.

    Parameters
    ----------
    dim : int
        Number of dimensions.
    n : int or list
        Number of points per dimension or a list with the number of points per dimension.
    box : list of lists
        List of lists containing the lower and upper bounds of the box.

    Returns
    -------
    x : numpy.ndarray
        Regular grid in the dim-dimensional hyperrectangle.
    """

    # Read argument 'n'
    if not isinstance(n, list):
        n = [n for i in range(dim)]

    # Read argument 'box'
    xmin, xmax = box[0], box[1]

    # levels
    levels = [np.linspace(xmin[i], xmax[i], n[i]) for i in range(dim)]

    # Construct a full factorial design x
    Xv = np.meshgrid(*levels, copy=True, sparse=False, indexing="ij")
    Xv = np.array(Xv)

    N = np.prod(n)
    x = np.zeros((N, dim))
    for i in range(dim):
        x[:, i] = Xv[i].reshape(
            N,
        )

    return x


def randunif(dim, n, box):
    """
    Generate a random uniform sample in the specified box.

    Parameters
    ----------
    dim : int
        Number of dimensions.
    n : int
        Number of points in the sample.
    box : list of lists
        List of lists containing the lower and upper bounds of the box.

    Returns
    -------
    numpy.ndarray
        Random uniform sample in the specified box.
    """
    sample = np.random.rand(n, dim)
    sample = scale(sample, box)

    return sample


def ldrandunif(dim, n, box, max_iter=50):
    """
    Generate a low discrepancy random uniform sample in the specified box.

    Parameters
    ----------
    dim : int
        Number of dimensions.
    n : int
        Number of points in the sample.
    box : list of lists
        List of lists containing the lower and upper bounds of the box.
    max_iter : int, optional
        Maximum number of iterations for optimization, default is 50.

    Returns
    -------
    numpy.ndarray
        Low discrepancy random uniform sample in the specified box.

    Notes
    -----
    FIXME: optimization method
    """
    mindiscrepany = 1e6  # large number
    for i in range(max_iter):
        sample = np.random.rand(n, dim)
        d = discrepancy(sample)
        if d < mindiscrepany:
            mindiscrepany = d
            sample_ld = sample

    sample_ld = scale(sample_ld, box)

    return sample_ld


def maximinlhs(dim, n, box, max_iter=1000):
    """
    Generate a maximin Latin Hypercube Sample (LHS) within the specified box.

    Parameters
    ----------
    dim : int
        Number of dimensions.
    n : int
        Number of points in the sample.
    box : list of lists
        List of lists containing the lower and upper bounds of the box.
    max_iter : int, optional
        Maximum number of iterations for finding the sample with the maximum minimum distance, default is 1000.

    Returns
    -------
    numpy.ndarray
        Maximin Latin Hypercube Sample within the specified box.
    """
    sampler = qmc.LatinHypercube(d=dim, optimization=None)

    maximindist = 0
    for i in range(max_iter):
        sample = sampler.random(n)
        d = mindist(sample)
        if d > maximindist:
            maximindist = d
            sample_maximin = sample

    sample_maximin = scale(sample_maximin, box)

    return sample_maximin


def maximinldlhs(dim, n, box):
    """
    Generate a maximin low-discrepancy Latin Hypercube Sample (LHS) within the specified box.

    Parameters
    ----------
    dim : int
        Number of dimensions.
    n : int
        Number of points in the sample.
    box : list of lists
        List of lists containing the lower and upper bounds of the box.

    Returns
    -------
    numpy.ndarray
        Maximin low-discrepancy Latin Hypercube Sample within the specified box.
    """
    sampler = qmc.LatinHypercube(d=dim, optimization="random-cd")

    max_iter = 10
    maximindist = 0
    for i in range(max_iter):
        sample = sampler.random(n)
        d = mindist(sample)
        if d > maximindist:
            maximindist = d
            sample_maximin = sample

    sample_maximin = scale(sample_maximin, box)

    return sample_maximin



"""
GP Conditional Sample Paths

This script demonstrates Gaussian Process (GP) regression with conditional sample paths
on heteroscedastic noisy data. It generates synthetic data using a predefined test function (two bumps)
and adds heteroscedastic Gaussian noise to the observed points. The script then constructs a GP
model using a Matern kernel and a constant mean function. It generates sample paths from the GP
prior, and then generates conditional sample paths given the observed data.

Copyright (c) 2022-2023, CentraleSupelec
Author: Emmanuel Vazquez <emmanuel.vazquez@centralesupelec.fr>
License: GPLv3 (see LICENSE)
"""

import math
import numpy as np
import gpmp.num as gnp
import gpmp as gp
import gpmp.misc.plotutils as plotutils


def generate_data():
    dim = 1
    box = [[-1], [1]]
    
    nt = 200
    # build a regular grid and evaluate the test function
    xt_ = gp.misc.designs.regulargrid(dim, nt, box)
    zt = gp.misc.testfunctions.twobumps(xt_)

    # extend xt so that the second column is the noise variance
    xt = np.hstack((xt_, np.zeros((nt, 1))))

    # observations
    noise_std_func = lambda x : 0.1 + (x + 1)**2
    
    xi1 = gp.misc.designs.regulargrid(dim, 30, box)
    xi2 = gp.misc.designs.regulargrid(dim, 50, [[0], [1]])
    xi_union = np.vstack((xi1, xi2))
    ni = xi_union.shape[0]

    # Calculate the noise standard deviation for each observation point
    noise_std = noise_std_func(xi_union)

    # Add the noise variance (noise_std**2) as the last column of xi
    xi = np.hstack((xi_union, noise_std**2))

    # Evaluation results with heteroscedastic Gaussian noise
    u = np.random.normal(size=(xi.shape[0], 1))
    zi = gp.misc.testfunctions.twobumps(xi_union).reshape((-1, 1)) + noise_std * u

    return xt, zt, xi, zi


def kernel_ii_or_tt(x, param, pairwise=False):
    """Covariance of the observations at points given by x
    Parameters
    ----------
    x : ndarray(n, d + 1)
        Data points in dimension d. The last column is the noise
        variance at the location.
    param : ndarray(1 + d)
        sigma2 and range parameters

    """
    p = 2
    sigma2 = gnp.exp(param[0])
    loginvrho = param[1]
    noise_variance = x[:, -1]  # Extract noise variance from the last column of x
    noise_variance = gnp.asarray(noise_variance)

    if pairwise:
        # return a vector of covariances
        K = sigma2 * gnp.ones((x.shape[0], )) + x[:, -1]
    else:
        # return a covariance matrix between observations
        K = gnp.scaled_distance(loginvrho, x[:, :-1], x[:, :-1])
        K = sigma2 * gp.kernel.maternp_kernel(p, K) + gnp.diag(noise_variance) 

    return K


def kernel_it(x, y, param, pairwise=False):
    """Covariance between observations and prediction points
    """
    p = 2
    sigma2 = gnp.exp(param[0])
    loginvrho = param[1]

    if pairwise:
        # return a vector of covariances
        K = gnp.scaled_distance_elementwise(loginvrho, x[:, :-1], y[:, :-1])
    else:
        # return a covariance matrix
        K = gnp.scaled_distance(loginvrho, x[:, :-1], y[:, :-1])

    K = sigma2 * gp.kernel.maternp_kernel(p, K)
    return K


def kernel(x, y, param, pairwise=False):
    if y is x or y is None:
        return kernel_ii_or_tt(x, param, pairwise)
    else:
        return kernel_it(x, y, param, pairwise)


def constant_mean(x, param):
    return gnp.ones((x.shape[0], 1))


def visualize(xt, zt, xi, zi, zpm, zpv, zsim, zpsim):
    fig = plotutils.Figure(isinteractive=True)
    fig.plot(xt[:, 0], zt, 'C0', linestyle=(0, (5, 5)), linewidth=1)
    fig.plot(xi[:, 0], zi, 'rs')
    fig.plotgp(xt[:, 0], zpm, zpv)
    fig.xylabels('x', 'z')
    fig.title('Data and Prediction, no sample paths')
    fig.show(grid=True)

    ni = xi.shape[0]
    nt = xt.shape[0]
    xixt = np.vstack((xi, xt))
    xi_ind = np.arange(ni)
    xt_ind = np.arange(nt) + ni

    fig = plotutils.Figure(isinteractive=True)
    fig.plot(xixt[xt_ind, 0], zsim[xt_ind])
    fig.ax.set_prop_cycle(None)
    fig.plot(xixt[xi_ind, 0], zsim[xi_ind], 'o')
    fig.title('Unconditional sample paths, with simulated noisy observations for each sample path')
    fig.show(grid=True)

    fig = plotutils.Figure(isinteractive=True)
    fig.plot(xixt[xt_ind, 0], zt, 'C2', linewidth=1)
    fig.plot(xixt[xt_ind, 0], zpsim, 'C0', linewidth=1)
    fig.plot(xi[:, 0], zi, 'rs')
    fig.plotgp(xt[:, 0], zpm, zpv)
    fig.title('Conditional Sample Paths')
    fig.show(grid=True)


def main():
    xt, zt, xi, zi = generate_data()

    mean = constant_mean
    meanparam = None
    covparam = gnp.array([math.log(0.5**2), math.log(1 / .7)])
    model = gp.core.Model(mean, kernel, meanparam, covparam)

    zpm, zpv, lambda_t = model.predict(xi, zi, xt, return_lambdas=True)

    ni = xi.shape[0]
    nt = xt.shape[0]
    xixt = np.vstack((xi, xt))
    xi_ind = np.arange(ni)
    xt_ind = np.arange(nt) + ni

    n_samplepaths = 3
    zsim = model.sample_paths(xixt, n_samplepaths, method='svd')
    zpsim = model.conditional_sample_paths(zsim, xi_ind, zi, xt_ind, lambda_t)

    visualize(xt, zt, xi, zi, zpm, zpv, zsim, zpsim)


if __name__ == "__main__":
    main()



# --------------------------------------------------------------
# Author: Emmanuel Vazquez <emmanuel.vazquez@centralesupelec.fr>
# Copyright (c) 2023, CentraleSupelec
# License: GPLv3 (see LICENSE)
# --------------------------------------------------------------
import time
import math
import numpy as np
import gpmp.num as gnp
import gpmp as gp
from gpmp.misc.dataframe import DataFrame, ftos
import matplotlib.pyplot as plt


def diag(model, info_select_parameters, xi, zi):
    """Run model diagnosis and display the results.

    Parameters
    ----------
    model : object
        GP Model object.
    info_select_parameters : object
        Information object containing the parameter selection process.
    xi : array-like
        Input data matrix.
    zi : array-like
        Output data matrix.
    """
    md = modeldiagnosis_init(model, info_select_parameters)
    model_diagnosis_disp(md, xi, zi)


def perf(model, xi, zi, loo=True, loo_res=None, xtzt=None, zpmzpv=None):
    perf = compute_performance(model, xi, zi, loo, loo_res, xtzt, zpmzpv)
    perf_disp = perf
    try:
        perf_disp.pop("loo_pit")
    except:
        pass
    try:
        perf_disp.pop("test_pit")
    except:
        pass
    print("Prediction performances")
    print("-----------------------")
    pretty_print_dictionnary(perf_disp)


def modeldiagnosis_init(model, info):
    """Build model diagnosis based on the provided model and information.

    Parameters
    ----------
    model : object
        Model object.
    info : object
        Information object containing the parameter selection process.

    Returns
    -------
    dict
        Model diagnosis information.
    """
    md = {
        "optim_info": info,
        "param_selection": {},
        "parameters": {},
        "loo": {},
        "data": {},
    }

    md["param_selection"] = {
        "cvg_reached": info.success,
        "optimal_val": info.best_value_returned,
        "n_evals": info.nfev,
        "time": info.total_time,
        "initial_val": float(info.selection_criterion(info.initial_params)),
        "final_val": info.fun,
    }

    covparam = gnp.asarray(model.covparam)
    md["parameters"] = sigma_rho_from_covparam(covparam)

    return md


def compute_performance(model, xi, zi, loo=True, loo_res=None, xtzt=None, zpmzpv=None):
    """Compute performance metrics of the GP model.

    Parameters
    ----------
    model : instance of GP model
        The GP model used to make predictions.
    xi : ndarray of shape (n, d)
        The input data used to fit the GP model.
    zi : ndarray of shape (n,)
        The target data used to fit the GP model.
    loo : bool, optional
        Whether or not to compute the leave-one-out (LOO) metrics. Default is True.
    loo_res : tuple, optional
        The output of the `loo` method of the GP model if already computed. Default is None.
    xtzt : tuple, optional
        The test input data and corresponding targets to be used for computing test set metrics. Default is None.
    zpmzpv : tuple, optional
        The predicted mean and variance on the test set data. Default is None.

    Returns
    -------
    perf : dict

        A dictionary containing the computed performance metrics. The
        metrics that are computed depend on the input arguments.
        If `loo` is True, the following keys are present:
        - 'data_tss': total sum of squares in the data.
        - 'loo_press': predictive residual error sum of squares in the
          leave-one-out predictions.
        - 'loo_Q2': coefficient of determination for the leave-one-out predictions.
        - 'loo_log10ratio': logarithm of the predictive
          residual error sum of squares to the total sum of squares in
          the leave-one-out predictions.
        - 'loo_pit': probability integral transform (PIT) of the
          leave-one-out predictions.  If `xtzt` and `zpmzpv` are not
          None, the following keys are present:
        - 'test_tss': total sum of squares in the test set predictions.
        - 'test_press': predictive residual error sum of squares in
          the test set predictions.
        - 'test_Q2': coefficient of determination for the test set predictions.
        - 'test_log10ratio': logarithm of the predictive
          residual error sum of squares to the total sum of squares in
          the test set predictions.
        - 'test_pit': PIT of the test set predictions.
    """
    xi = gnp.asarray(xi)
    zi = gnp.asarray(zi)

    if loo and loo_res == None:
        zloom, zloov, eloo = model.loo(xi, zi)
    elif loo and type(loo_res) is tuple:
        zloom, zloov, eloo = loo_res

    test_set = False
    if xtzt is not None:
        test_set = True
        xt, zt = xtzt
        xt = gnp.asarray(xt)
        zt = gnp.asarray(zt)
    if test_set and zpmzpv is None:
        zpm, zpv = model.predict(xi, zi, xt)
    elif xtzt is not None and zpmzpv is not None:
        zpm, zpv = zpmzpv
        zpm = gnp.asarray(zpm)
        zpv = gnp.asarray(zpv)

    perf = {}

    if loo:
        # total sum of squares
        perf["data_tss"] = gnp.norm(zi - gnp.mean(zi), ord=2)
        # Predictive residual Error sum of squares
        perf["loo_press"] = gnp.norm(eloo, ord=2)
        # Q2
        perf["loo_Q2"] = 1 - perf["loo_press"] / perf["data_tss"]
        # L2err
        perf["loo_log10ratio"] = gnp.log10(perf["loo_press"] / perf["data_tss"])
        # PIT
        perf["loo_pit"] = gnp.normal.cdf(zi, loc=zloom, scale=gnp.sqrt(zloov))

    if test_set:
        perf["test_tss"] = gnp.norm(zt - gnp.mean(zt), ord=2)
        perf["test_press"] = gnp.norm(zpm - zt, ord=2)
        perf["test_Q2"] = 1 - perf["test_press"] / perf["test_tss"]
        perf["test_log10ratio"] = gnp.log10(perf["test_press"] / perf["test_tss"])
        perf["test_pit"] = gnp.normal.cdf(zt, loc=zpm, scale=gnp.sqrt(zpv))

    return perf


def plot_pit_ecdf(pit, fig=None):
    """Plot the empirical cumulative distribution function (ECDF) of a Probability
    Integral Transform (PIT) vector.

    Parameters
    ----------
    pit : gnp.ndarray, shape (n,)
        An array of PIT values.
    fig : matplotlib.figure.Figure, optional
        Matplotlib figure object to plot the PIT ECDF on. If None, a
        new figure is created. Default is None.

    Returns
    -------
    None
    """
    n = pit.shape[0]
    p = gnp.concatenate((gnp.array([0]), gnp.linspace(0, 1, n)))
    pit_sorted = gnp.concatenate((gnp.array([0.0]), gnp.sort(pit)))

    if fig is None:
        plt.figure()
    plt.step(pit_sorted, p)
    plt.plot([0, 1], [0, 1])
    plt.title("PIT (Probability Integral Transform) ECDF")
    plt.show()


def model_diagnosis_disp(md, xi, zi):
    """Display model diagnosis information.

    Parameters
    ----------
    md : dict
        Model diagnosis information.
    xi : array-like
        Input data matrix.
    zi : array-like
        Output data matrix.
    """
    print("Model diagnosis")
    print("----------------")
    print("  ***  Parameter selection")
    pretty_print_dictionnary(md["param_selection"])

    print("  ***  Parameters")
    pretty_print_dictionnary(md["parameters"])

    print("  ***  Data")
    print("   {:>0}: {:d}".format("count", zi.shape[0]))
    print("   ----")

    param_values = np.array(list(md["parameters"].values()))

    # zi
    if zi.ndim == 1:
        rownames = ["zi"]
    else:
        rownames = [f"zi_{j}" for j in range(zi.shape[1])]

    df_zi = describe_array(zi, rownames, 1 / param_values[0])

    # xi
    rownames = [f"xi_{j}" for j in range(xi.shape[1])]
    df_xi = describe_array(xi, rownames, 1 / param_values[1:])

    # zi + xi
    print(df_zi.concat(df_xi))


def plot_2d_criterion_params(
    model,
    info,
    param_indices=(0, 1),
    param_names=None,
    criterion_name='selection criterion',
):
    """
    Plot selection criterion profile for any two parameters in covparam.

    Parameters
    ----------
    model : object
        Model object.
    info : object
        Information object containing the parameter selection process.
    param_indices : tuple, optional
        Indices of the two parameters to plot (default is (0, 1)).
    param_names : list, optional
        Names of the two parameters for labeling the axes (default is None).
    criterion_name : string, optional
        Name of the selection criterion to be displayed in the title
    """
    n = 180
    tic = time.time()
    
    def print_progress(i):
        elapsed_time = time.time() - tic
        average_time_per_iteration = elapsed_time / (i + 1)
        remaining_time = average_time_per_iteration * (n - (i + 1))
        percentage = (i + 1) / n * 100
        print(
            f"       Progress: {percentage:.2f}% | time remaining: {remaining_time:.1f}s",
            end="\r",
        )
    def print_final_time():
        elapsed_time = time.time() - tic
        print(f"       Progress: 100% complete | Total time: {elapsed_time:.3f}s")
        print(f"       number of evaluations: {n * n}")

    print(f"  ***  Computing {criterion_name} profile for plotting...")
        
    param_1_idx, param_2_idx = param_indices

    # Initialize param1 and param2 based on their indices (standard deviation or scale parameter)
    param_1_0 = math.exp(
        model.covparam[param_1_idx] / 2
        if param_1_idx == 0
        else -model.covparam[param_1_idx]
    )
    param_2_0 = math.exp(
        model.covparam[param_2_idx] / 2
        if param_2_idx == 0
        else -model.covparam[param_2_idx]
    )

    f = 4  # multiplying factor
    param_1 = np.logspace(
        math.log10(param_1_0) - math.log10(f), math.log10(param_1_0) + math.log(f), n
    )
    param_2 = np.logspace(
        math.log10(param_2_0) - math.log10(f), math.log10(param_2_0) + math.log(f), n
    )

    param_1_mesh, param_2_mesh = np.meshgrid(param_1, param_2)
    log_param_1 = (
        np.log(param_1_mesh**2) if param_1_idx == 0 else np.log(1 / param_1_mesh)
    )
    log_param_2 = (
        np.log(param_2_mesh**2) if param_2_idx == 0 else np.log(1 / param_2_mesh)
    )

    selection_criterion = info.selection_criterion
    selection_criterion_values = np.zeros((n, n))

    covparam = gnp.copy(info.covparam)
    for i in range(n):
        print_progress(i)
        for j in range(n):
            covparam = gnp.set_elem1(covparam, param_1_idx, log_param_1[i, j])
            covparam = gnp.set_elem1(covparam, param_2_idx, log_param_2[i, j])
            selection_criterion_values[i, j] = selection_criterion(covparam)

    selection_criterion_values = np.nan_to_num(selection_criterion_values, copy=False)
    print_final_time()

    shift_criterion = True
    shift = -np.min(selection_criterion_values) if shift_criterion else 0

    # Plot the selection criterion profile
    plt.figure()
    plt.contourf(
        np.log10(param_1_mesh),
        np.log10(param_2_mesh),
        np.log10(np.maximum(1e-2, selection_criterion_values + shift)),
    )
    plt.plot(
        0.5 * np.log10(np.exp(info.covparam[param_1_idx])),
        -np.log10(np.exp(info.covparam[param_2_idx])),
        "ro",
    )
    plt.plot(
        0.5 * np.log10(np.exp(info.covparam0[param_1_idx])),
        -np.log10(np.exp(info.covparam0[param_2_idx])),
        "bo",
    )

    # Define axis labels (use names if provided)
    x_label = param_names[0] if param_names else f"Parameter {param_1_idx} (log10)"
    y_label = param_names[1] if param_names else f"Parameter {param_2_idx} (log10)"

    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.title(f'log10 of the {"shifted " if shift_criterion else ""}{criterion_name}')
    plt.colorbar()
    plt.show()


def plot_likelihood_sigma_rho(model, info, criterion_name="negative log restricted likelihood"):
    """
    Specific case of likelihood plotting for sigma (param 0) and rho (param 1).

    Parameters
    ----------
    model : object
        Model object.
    info : object
        Information object containing the parameter selection process.
    criterion_name : string, optional
        Name of the selection criterion to be displayed in the title
    """
    # Use the generalized function to plot sigma and rho with custom names
    plot_2d_criterion_params(
        model,
        info,
        param_indices=(0, 1),
        param_names=["sigma (log10)", "rho (log10)"],
        criterion_name=criterion_name,
    )


def sigma_rho_from_covparam(covparam):
    """Extract sigma and rho parameters from the covariance parameters.

    Parameters
    ----------
    covparam : array-like
        Covariance parameters.

    Returns
    -------
    dict
        Dictionary containing sigma and rho values.
    """
    pdict = {}
    pdict["sigma"] = gnp.exp(0.5 * covparam[0])
    for i in range(covparam.shape[0] - 1):
        k = "rho{:d}".format(i)
        v = gnp.exp(-covparam[i + 1])
        pdict[k] = v

    return pdict


def describe_array(x, rownames, normalizing_factor=None):
    """Create a DataFrame containing descriptive statistics for the given data.

    Parameters
    ----------
    x : array-like
        Input data matrix.
    rownames : list
        List of row names for the DataFrame.
    normalizing_factor : float, optional
        Normalizing factor to compute the 'delta_norm' column, by default None.

    Returns
    -------
    DataFrame
        DataFrame with descriptive statistics.
    """
    x = np.array(x)
    if normalizing_factor is None:
        n_descriptors = 5
        colnames = ["mean", "std", "min", "max", "delta"]
    else:
        n_descriptors = 6
        colnames = ["mean", "std", "min", "max", "delta", "delta_norm"]
    dim = 1 if x.ndim == 1 else x.shape[1]

    data = np.empty((dim, n_descriptors))

    data[:, 0] = np.mean(x, axis=0)
    data[:, 1] = np.std(x, axis=0)
    data[:, 2] = np.min(x, axis=0)
    data[:, 3] = np.max(x, axis=0)
    data[:, 4] = data[:, 3] - data[:, 2]

    if normalizing_factor is not None:
        data[:, 5] = data[:, 4] * normalizing_factor

    return DataFrame(data, colnames, rownames)


def pretty_print_dictionnary(d, fp=4):
    """Print a dictionary with formatted values.

    Parameters
    ----------
    d : dict
        The dictionary to be printed.
    fp : int, optional
        Number of decimal places for floating-point values, by default 4.
    """
    max_key_length = max(15, max(len(str(k)) for k in d.keys()) + 2)

    for k, v in d.items():
        if not gnp.isscalar(v):
            v = v.item()
        if isinstance(v, float):
            s = f"{{:>{max_key_length}s}}: {{:s}}"
            print(s.format(k, ftos(v, fp)))
        else:
            print(f"{k:>{max_key_length}s}: {v}")



"""Prediction of some classical test functions in dimension > 2

An anisotropic Matern covariance function is used for the Gaussian
Process (GP) prior. The parameters of this covariance function
(variance and ranges) are estimated using the Restricted Maximum
Likelihood (ReML).

The mean function of the GP prior is assumed to be constant and
unknown.

The function is sampled on a space-filling Latin Hypercube design, and
the data is assumed to be noiseless.

----
Author: Emmanuel Vazquez <emmanuel.vazquez@centralesupelec.fr>
Copyright (c) 2022-2024, CentraleSupelec
License: GPLv3 (see LICENSE)
"""
import gpmp.num as gnp
import gpmp as gp
import matplotlib.pyplot as plt


def choose_test_case(problem):
    if problem == 1:
        problem_name = "Hartmann4"
        f = gp.misc.testfunctions.hartmann4
        dim = 4
        box = [[0.0] * 4, [1.0] * 4]
        ni = 40
        xi = gp.misc.designs.ldrandunif(dim, ni, box)
        nt = 1000
        xt = gp.misc.designs.ldrandunif(dim, nt, box)

    elif problem == 2:
        problem_name = "Hartmann6"
        f = gp.misc.testfunctions.hartmann6
        dim = 6
        box = [[0.0] * 6, [1.0] * 6]
        ni = 500
        xi = gp.misc.designs.ldrandunif(dim, ni, box)
        nt = 1000
        xt = gp.misc.designs.ldrandunif(dim, nt, box)

    elif problem == 3:
        problem_name = "Borehole"
        f = gp.misc.testfunctions.borehole
        dim = 8
        box = [
            [0.05, 100.0, 63070.0, 990.0, 63.1, 700.0, 1120.0, 9855.0],
            [0.15, 50000.0, 115600.0, 1110.0, 116.0, 820.0, 1680.0, 12045.0],
        ]
        ni = 30
        xi = gp.misc.designs.maximinldlhs(dim, ni, box)
        nt = 1000
        xt = gp.misc.designs.ldrandunif(dim, nt, box)

    elif problem == 4:
        problem_name = "detpep8d"
        f = gp.misc.testfunctions.detpep8d
        dim = 8
        box = [[0.0] * 8, [1.0] * 8]
        ni = 60
        xi = gp.misc.designs.maximinldlhs(dim, ni, box)
        nt = 1000
        xt = gp.misc.designs.ldrandunif(dim, nt, box)

    return problem_name, f, dim, box, ni, xi, nt, xt


def constant_mean(x, param):
    return gnp.ones((x.shape[0], 1))


def kernel(x, y, covparam, pairwise=False):
    p = 10
    return gp.kernel.maternp_covariance(x, y, p, covparam, pairwise)


def visualize_predictions(problem_name, zt, zpm):
    plt.figure()
    plt.plot(zt, zpm, "ko")
    (xmin, xmax), (ymin, ymax) = plt.xlim(), plt.ylim()
    xmin = min(xmin, ymin)
    xmax = max(xmax, ymax)
    plt.plot([xmin, xmax], [xmin, xmax], "--")
    plt.title(problem_name)
    plt.show()


def main():
    problem = 1
    problem_name, f, dim, box, ni, xi, nt, xt = choose_test_case(problem)

    zi = f(xi)
    zt = f(xt)

    model = gp.core.Model(constant_mean, kernel)

    model, info = gp.kernel.select_parameters_with_reml(model, xi, zi, info=True)
    gp.misc.modeldiagnosis.diag(model, info, xi, zi)

    (zpm, zpv) = model.predict(xi, zi, xt)

    visualize_predictions(problem_name, zt, zpm)

    zloom, zloov, eloo = model.loo(xi, zi)
    gp.misc.plotutils.plot_loo(zi, zloom, zloov)

    gp.misc.plotutils.crosssections(model, xi, zi, box, [0, 1], list(range(dim)))

    print('')
    
    gp.misc.modeldiagnosis.perf(
        model,
        xi,
        zi,
        loo=True,
        loo_res=(zloom, zloov, eloo),
        xtzt=(xt, zt),
        zpmzpv=(zpm, zpv),
    )


if __name__ == "__main__":
    main()



"""
Gaussian processes in 2D

An anisotropic Matern covariance function is used for the Gaussian
Process (GP) prior. The parameters of this covariance function
(variance and ranges) are estimated using the Restricted Maximum
Likelihood (ReML) method.

The mean function of the GP prior is assumed to be constant and
unknown.

The function is sampled on a space-filling Latin Hypercube design, and
the data is assumed to be noiseless.

----
Author: Emmanuel Vazquez <emmanuel.vazquez@centralesupelec.fr>
Copyright (c) 2022-2024, CentraleSupelec
License: GPLv3 (see LICENSE)
----
This example is based on the file stk_example_kb03.m from the STK at
https://github.com/stk-kriging/stk/ 
by Julien Bect and Emmanuel Vazquez, released under the GPLv3 license.

Original copyright notice:

   Copyright (c) 2015, 2016, 2018 CentraleSupelec
   Copyright (c) 2011-2014 SUPELEC
----
"""

import numpy as np
import gpmp.num as gnp
import gpmp as gp
import matplotlib.pyplot as plt


# Test function selection
def select_test_function(case_num):
    if case_num == 1:
        f = gp.misc.testfunctions.braninhoo
        dim = 2
        box = [[-5, 0], [10, 15]]
        ni = 20
    elif case_num == 2:
        f = gp.misc.testfunctions.wave
        dim = 2
        box = [[-1, -1], [1, 1]]
        ni = 40
    return f, dim, box, ni


def create_model():
    def constant_mean(x, param):
        return gnp.ones((x.shape[0], 1))

    def kernel(x, y, covparam, pairwise=False):
        p = 6
        return gp.kernel.maternp_covariance(x, y, p, covparam, pairwise)

    return gp.core.Model(constant_mean, kernel)


def main():
    case_num = 2
    f, dim, box, ni = select_test_function(case_num)

    # Compute the function on a 80 x 80 regular grid
    nt = [80, 80]
    xt = gp.misc.designs.regulargrid(dim, nt, box)
    zt = f(xt)

    design_type = 'ld'
    if design_type == 'lhs':
        xi = gp.misc.designs.maximinlhs(dim, ni, box)
    elif design_type == 'ld':
        xi = gp.misc.designs.ldrandunif(dim, ni, box)
    zi = f(xi)

    model = create_model()

    # Parameter selection
    covparam0 = gp.kernel.anisotropic_parameters_initial_guess(model, xi, zi)
    nlrl, dnlrl = gp.kernel.make_selection_criterion_with_gradient(
        model,
        gp.kernel.negative_log_restricted_likelihood,
        xi,
        zi)
    covparam_reml, info = gp.kernel.autoselect_parameters(covparam0, nlrl, dnlrl, info=True)
    model.covparam = gnp.asarray(covparam_reml)
    gp.misc.modeldiagnosis.diag(model, info, xi, zi)

    # Prediction
    (zpm, zpv) = model.predict(xi, zi, xt)

    # Visualization
    cmap = plt.get_cmap('PiYG')
    contour_lines = 30

    fig, axes = plt.subplots(nrows=2, ncols=2)
    data = [zt, zpm, np.abs(zpm - zt), np.sqrt(zpv)]
    titles = [
        'function to be approximated',
        f'approximation from {ni} points',
        'true approx error',
        'posterior std'
    ]

    for ax, z, title in zip(axes.flat, data, titles):
        cs = ax.contourf(xt[:, 0].reshape(nt),
                         xt[:, 1].reshape(nt),
                         z.reshape(nt),
                         levels=contour_lines,
                         cmap=cmap)
        ax.plot(xi[:, 0], xi[:, 1], 'ro', label='data')
        ax.set_title(title)
        ax.set_xlabel('$x_1$')
        ax.set_ylabel('$x_2$')
        ax.legend()
        fig.colorbar(cs, ax=ax, shrink=0.9)

    plt.show()

    # Predictions vs truth
    plt.figure()
    plt.plot(zt, zpm, 'ko')
    (xmin, xmax), (ymin, ymax) = plt.xlim(), plt.ylim()
    xmin = min(xmin, ymin)
    xmax = max(xmax, ymax)
    plt.plot([xmin, xmax], [xmin, xmax], '--')
    plt.xlabel('true values')
    plt.ylabel('predictions')
    plt.show()

    # LOO predictions
    zloom, zloov, eloo = model.loo(xi, zi)
    gp.misc.plotutils.plot_loo(zi, zloom, zloov)

    gp.misc.plotutils.crosssections(model, xi, zi, box, [0, 20], [0, 1])

if __name__ == '__main__':
    main()



## --------------------------------------------------------------
# Author: Emmanuel Vazquez <emmanuel.vazquez@centralesupelec.fr>
# Copyright (c) 2022-2024, CentraleSupelec
# License: GPLv3 (see LICENSE)
## --------------------------------------------------------------
import time
import numpy as np
import gpmp.num as gnp
from scipy.optimize import minimize, OptimizeWarning
from math import log, sqrt


## -- kernels


def exponential_kernel(h):
    """Exponential kernel.

    The exponential kernel is defined as:

    .. math::

        k(h) = \exp(-h)

    where `h` represents the distances between points.

    Parameters
    ----------
    h : gnp.array, shape (n,)
        An array of distances between points.

    Returns
    -------
    gnp.array, shape (n,)
        An array of the exponential kernel values corresponding to the input distances.
    """
    return gnp.exp(-h)


def matern32_kernel(h):
    """Matérn 3/2 kernel.

    The Matérn 3/2 kernel is defined as:

    .. math::

        K(h) = (1 + 2\\sqrt{3/2}h) \\exp(-2\\sqrt{3/2}h)

    where `h` represents the distances between points.

    Parameters
    ----------
    h : gnp.array, shape (n,)
        An array of distances between points.

    Returns
    -------
    gnp.array, shape (n,)
        An array of the Matérn 3/2 kernel values corresponding to the input distances.
    """
    nu = 3.0 / 2.0
    c = 2.0 * sqrt(nu)
    t = c * h

    return (1.0 + t) * gnp.exp(-t)


def compute_gammaln(up_to_p):
    """Compute gammaln values."""
    return [gnp.asarray(gnp.gammaln(i)) for i in range(2 * up_to_p + 2)]


gln = []
pmax = -1


def maternp_kernel(p: int, h):
    """Matérn kernel with half-integer regularity nu = p + 1/2.

    The Matérn kernel is defined as in Stein 1999, page 50:

    .. math::

        K(h) = \\frac{1}{\\Gamma(\\nu) 2^{\\nu - 1}} (\\sqrt{2 \\nu} h)^\\nu K_\\nu(\\sqrt{2 \\nu} h)

    Where:
    - :math:`h` represents the distances between points.
    - :math:`\\nu` is the regularity of the kernel.
    - :math:`K_\\nu` is the modified Bessel function of the second kind of order :math:`\\nu`.

    The implementation provided in this function uses this half-integer simplification.

    In the particular case of half-integer regularity (nu = p + 1/2), the Matérn kernel
    simplifies to a product of an exponential term and a polynomial term (Watson 1922,
    A treatise on the theory of Bessel functions, pp. 80, Abramowitz and Stegun 1965, pp. 374-379, 443-444):

    .. math::

        K(h) = \\exp(-2\\sqrt{\\nu}h) \\frac{\\Gamma(p+1)}{\\Gamma(2p+1)} \\sum_{i=0}^{p} \\frac{(p+i)!}{i!(p-i)!} (4\\sqrt{\\nu}h)^{p-i}
    """
    global gln, pmax

    # Check if p exceeds pmax and compute gammaln cache if needed
    if p > pmax:
        gln = compute_gammaln(p)
        pmax = p

    h = gnp.inftobigf(h)
    c = 2.0 * sqrt(p + 0.5)
    twoch = 2.0 * c * h
    polynomial = gnp.ones(h.shape)

    for i in range(p):
        exp_log_combination = gnp.exp(
            gln[p + 1] - gln[2 * p + 1] + gln[p + i + 1] - gln[i + 1] - gln[p - i + 1]
        )
        twoch_pow = twoch ** (p - i)
        polynomial += exp_log_combination * twoch_pow

    return gnp.exp(-c * h) * polynomial


def maternp_covariance_ii_or_tt(x, p, param, pairwise=False):
    """Covariance between observations or predictands at x.

    The covariance matrix is computed using the Matérn kernel with half-integer regularity:

    .. math::

        K_{ij} = \\sigma^2  K(h_{ij}) + \\epsilon \\delta_{ij}

    where `K(h_{ij})` is the Matérn kernel value for the distance `h_{ij}` between points `x_i` and `x_j`,
    `sigma^2` is the variance, `delta_{ij}` is the Kronecker delta, and `epsilon` is a small positive constant.

    Parameters
    ----------
    x : gnp.array, shape (nx, d)
        Observation points.
    p : int
        Half-integer regularity nu = p + 1/2.
    param : gnp.array, shape (1 + d,)
        sigma2 and range parameters.
    pairwise : bool, optional
        Whether to return a covariance matrix k(x_i, x_j),
        for i and j = 1 ... nx, if pairwise is False, or a covariance
        vector k(x_i, x_i) if pairwise is True. Default is False.

    Returns
    -------
    K : gnp.array
        Covariance matrix (nx, nx) or covariance vector if pairwise is True.
    """
    sigma2 = gnp.exp(param[0])
    loginvrho = param[1:]
    nugget = 10.0 * sigma2 * gnp.finfo(gnp.float64).eps

    if pairwise:
        # return a vector of covariances
        K = sigma2 * gnp.ones((x.shape[0],))  # nx x 0
    else:
        # return a covariance matrix
        K = gnp.scaled_distance(loginvrho, x, x)  # nx x nx
        K = sigma2 * maternp_kernel(p, K) + nugget * gnp.eye(K.shape[0])

    return K


def maternp_covariance_it(x, y, p, param, pairwise=False):
    """Covariance between observations and prediction points.

    Parameters
    ----------
    x : ndarray, shape (nx, d)
        Observation points.
    y : ndarray, shape (ny, d)
        Observation points.
    p : int
        Half-integer regularity nu = p + 1/2.
    param : ndarray, shape (1 + d,)
        log(sigma2) and log(1/range) parameters.
    pairwise : bool, optional
        Whether to return a covariance matrix k(x_i, y_j),
        for i in 1 ... nx and j in 1 ... ny, if pairwise is False,
        or a covariance vector k(x_i, y_i) if pairwise is True. Default is False.

    Returns
    -------
    K : ndarray
        Covariance matrix (nx, ny) or covariance vector if pairwise is True.
    """
    sigma2 = gnp.exp(param[0])
    loginvrho = param[1:]

    if pairwise:
        # return a vector of distances
        K = gnp.scaled_distance_elementwise(loginvrho, x, y)  # nx x 0
    else:
        # return a distance matrix
        K = gnp.scaled_distance(loginvrho, x, y)  # nx x ny

    K = sigma2 * maternp_kernel(p, K)

    return K


def maternp_covariance(x, y, p, param, pairwise=False):
    """Matérn covariance function with half-integer regularity nu = p + 1/2.

    The kernel is defined in terms of the Euclidean distance, between
    pairs of input points. For the Matérn kernel, the distance measure
    is scaled by a length scale parameter, which controls how much
    influence distant points have on each other. The kernel has two
    hyperparameters: the length scale and a smoothness parameter,
    which is typically an integer or half-integer value.

    Parameters
    ----------
    x : ndarray, shape (nx, d)
        Observation points.
    y : ndarray, shape (ny, d) or None
        Prediction points. If None, it is assumed that y is x.
    p : int
        Half-integer regularity nu = p + 1/2.
    param : ndarray, shape (1 + d)
        Covariance parameters
        [log(sigma2) log(1/rho_1) log(1/rho_2) ...].
    pairwise : bool, optional
        If True, return a covariance vector k(x_i, y_i). If False,
        return a covariance matrix k(x_i, y_j) for i in the range 1 to nx
        and j in the range 1 to ny. Default is False.

    Returns
    -------
    K : ndarray
        Covariance matrix (nx, ny) or covariance vector if pairwise is True.

    Notes
    -----
    An isotropic covariance is obtained if param = [log(sigma2) log(1/rho)]
    (only one length scale parameter).
    """
    if y is x or y is None:
        return maternp_covariance_ii_or_tt(x, p, param, pairwise)
    else:
        return maternp_covariance_it(x, y, p, param, pairwise)


## -- parameters


def anisotropic_parameters_initial_guess_zero_mean(model, xi, zi):
    """Anisotropic initialization strategy with zero mean.

    See anisotropic_parameters_initial_guess
    """
    xi_ = gnp.asarray(xi)
    zi_ = gnp.asarray(zi).reshape(-1, 1)
    n = xi_.shape[0]
    d = xi_.shape[1]

    delta = gnp.max(xi_, axis=0) - gnp.min(xi_, axis=0)
    rho = gnp.exp(gnp.gammaln(d / 2 + 1) / d) / (gnp.pi**0.5) * delta
    covparam = gnp.concatenate((gnp.array([log(1.0)]), -gnp.log(rho)))
    sigma2_GLS = 1.0 / n * model.norm_k_sqrd_with_zero_mean(xi_, zi_, covparam)

    return gnp.concatenate((gnp.log(sigma2_GLS), -gnp.log(rho)))


def anisotropic_parameters_initial_guess_constant_mean(model, xi, zi):
    """Anisotropic initialization strategy with a parameterized constant mean.

    This function provides initial parameter guesses for an
    anisotropic Gaussian process with a parameterized constant mean.

    Parameters
    ----------
    model : object
        The Gaussian process model object.
    xi : array_like, shape (n, d)
        Input data points used for fitting the GP model, where `n` is
        the number of points and `d` is the dimensionality.
    zi : array_like, shape (n, )
        Output (response) values corresponding to the input data points xi.

    Returns
    -------
    mean_GLS : float
        The generalized least squares (GLS) estimator of the
        mean. Computed as:

        .. math::

            m_{GLS} = \frac{\mathbf{1}^T K^{-1} \mathbf{z}}{\mathbf{1}^T K^{-1} \mathbf{1}}

    concatenated parameters : array_like
        An array containing the initialized :math:`\sigma^2_{GLS}` and :math:`\rho` values.
        The estimator :math:`\sigma^2_{GLS}` is given by:

        .. math::

            \sigma^2_{GLS} = \frac{1}{n} \mathbf{z}^T K^{-1} \mathbf{z}
    """
    xi_ = gnp.asarray(xi)
    zi_ = gnp.asarray(zi).reshape((-1, 1))  # Ensure zi_ is a column vector
    n = xi_.shape[0]
    d = xi_.shape[1]

    delta = gnp.max(xi_, axis=0) - gnp.min(xi_, axis=0)
    rho = gnp.exp(gnp.gammaln(d / 2 + 1) / d) / (gnp.pi**0.5) * delta

    covparam = gnp.concatenate((gnp.array([gnp.log(1.0)]), -gnp.log(rho)))
    zTKinvz, Kinv1, Kinvz = model.k_inverses(xi_, zi_, covparam)

    mean_GLS = gnp.sum(Kinvz) / gnp.sum(Kinv1)
    sigma2_GLS = (1.0 / n) * zTKinvz

    return mean_GLS.reshape(1), gnp.concatenate((gnp.log(sigma2_GLS), -gnp.log(rho)))


def anisotropic_parameters_initial_guess(model, xi, zi):
    """Anisotropic initialization strategy for parameters of a Gaussian process model.

    Given the observed data points and their values, this function
    computes an initial guess for the anisotropic parameters. The
    guess for :math:`\\sigma^2` is initialized using the Generalized
    Least Squares (GLS) estimate as described below.

    Parameters
    ----------
    model : object
        An instance of a Gaussian process model.
    xi : ndarray, shape (n, d)
        Locations of the observed data points.
    zi : ndarray, shape (n,)
        Observed values at the data points.

    Returns
    -------
    initial_params : ndarray
        Initial guess for the anisotropic parameters, comprising the
        estimate for :math:`\\sigma^2_{GLS}` followed by the estimates
        for the anisotropic lengthscales.

    Notes
    -----
    The GLS estimate for :math:`\\sigma^2` is given by:

    .. math::

        \\sigma^2_{GLS} = \\frac{1}{n} \\mathbf{z}^T \\mathbf{K}^{-1} \\mathbf{z}

    Where:

    * :math:`n` is the number of data points.
    * :math:`\\mathbf{z}` is the vector of observed data.
    * :math:`\\mathbf{K}` is the covariance matrix associated with the data locations.

    Additionally, the function uses a relation (not from the
    reference) between :math:`\\rho` and the volume of a ball in
    dimension :math:`d` for initialization:

    .. math::

        V_d(R) = \\frac{\\pi^{d/2} R^d}{\\Gamma(d/2+1)}

    Where :math:`R` is defined as :math:`\\rho / 2`.

    .. [1] Basak, S., Petit, S., Bect, J., & Vazquez, E. (2021).
       Numerical issues in maximum likelihood parameter estimation for
       Gaussian process interpolation. arXiv:2101.09747.
    """

    xi_ = gnp.asarray(xi)
    zi_ = gnp.asarray(zi).reshape(-1, 1)
    n = xi_.shape[0]
    d = xi_.shape[1]

    delta = gnp.max(xi_, axis=0) - gnp.min(xi_, axis=0)
    rho = gnp.exp(gnp.gammaln(d / 2 + 1) / d) / (gnp.pi**0.5) * delta

    covparam = gnp.concatenate((gnp.array([log(1.0)]), -gnp.log(rho)))
    sigma2_GLS = 1.0 / n * model.norm_k_sqrd(xi_, zi_, covparam)

    return gnp.concatenate((gnp.log(sigma2_GLS), -gnp.log(rho)))


def make_selection_criterion_with_gradient(
    model, selection_criterion, xi, zi, parameterized_mean=False, meanparam_len=1
):
    """Make selection criterion function with gradient.

    Parameters
    ----------
    model : object
        Instance of a Gaussian process model that needs parameter optimization.
    selection_criterion : function
        Selection criterion function. (See Notes.)
    xi : ndarray, shape (n, d)
        Locations of the observed data points.
    zi : ndarray, shape (n,)
        Observed values at the data points.
    parameterized_mean : bool, optional
        Whether to use mean parameter in the selection criterion.
    meanparam_len : int, optional
        Length of the mean parameter, used only if parameterized_mean is True.

    Returns
    -------
    crit_jit : function
        Selection criterion function with gradient.
    dcrit : function
        Gradient of the selection criterion function.

    Notes
    -----
    The `criterion` function should follow one of the following two forms:

    - For models without a parameterized mean: `criterion(model,
      covparam, xi, zi)` where `covparam` are the covariance
      parameters.

    - For models with a parameterized mean: `criterion(model,
      meanparam, covparam, xi, zi)` where both `meanparam` and
      `covparam` are passed.

    The function will automatically handle the form of the criterion
    based on whether a mean parameter is used (controlled by the
    `parameterized_mean` flag in the criterion definition).

    """
    xi_ = gnp.asarray(xi)
    zi_ = gnp.asarray(zi)

    if parameterized_mean:
        # make a selection criterion with mean parameter
        def crit_(param):
            meanparam = param[:meanparam_len]
            covparam = param[meanparam_len:]
            l = selection_criterion(model, meanparam, covparam, xi_, zi_)
            return l

    else:
        # make a selection criterion without mean parameter
        def crit_(covparam):
            l = selection_criterion(model, covparam, xi_, zi_)
            return l

    crit_jit = gnp.jax.jit(crit_)
    dcrit = gnp.jax.jit(gnp.grad(crit_jit))

    return crit_jit, dcrit


def autoselect_parameters(
    p0,
    criterion,
    gradient,
    bounds=None,
    silent=True,
    info=False,
    method="SLSQP",
    method_options={},
):
    """Optimize parameters using a provided criterion and gradient function.

    This function automatically optimizes the parameters of a given model based on
    a specified criterion function and its gradient. Different optimization methods
    can be used based on the `method` argument.

    Parameters
    ----------
    p0 : ndarray
        Initial guess of the parameters.
    criterion : function
        Function that computes the selection criterion for a given parameter set.
        It should take an ndarray of parameters and return a scalar value.
    gradient : function
        Function that computes the gradient of the selection criterion with respect
        to the parameters. It should take an ndarray of parameters and return an ndarray
        of the same shape.
    bounds : sequence of tuple, optional
        A sequence of (min, max) pairs specifying bounds for each parameter.
        Use None to indicate no bounds. Default is None.
    silent : bool, optional
        If True, suppresses optimization output messages. Default is True.
    info : bool, optional
        If True, returns additional information about the optimization process.
        Default is False.
    method : str, optional
        Optimization method to use. Supported methods are 'L-BFGS-B' and 'SLSQP'.
        Default is 'SLSQP'.
    method_options : dict, optional, default {}
        User options for the optimization method.

    Returns
    -------
    best : ndarray
        Array of optimized parameters.
    r : OptimizeResult, optional
        A dictionary of optimization information (only if `info=True`). This includes
        details like the initial parameters, final parameters, selection criterion function,
        and total time taken for optimization.

    Notes
    -----
    The function uses the `minimize` method from `scipy.optimize` for optimization.
    Depending on the backend (`gnp._gpmp_backend_`), different preparations are made
    for the criterion and gradient functions to ensure compatibility.
    """
    # Track optimization start time
    tic = time.time()

    # Setup to record parameter and criterion history
    history_params = []
    history_criterion = []
    best_criterion = float("inf")
    best_params = None

    def record_history(p, criterion_value):
        nonlocal best_criterion, best_params
        history_params.append(p.copy())
        history_criterion.append(criterion_value)

        if criterion_value < best_criterion:
            best_criterion = criterion_value
            best_params = p.copy()

    # Determine which backend to use and configure the criterion and gradient functions
    if gnp._gpmp_backend_ == "jax":
        # scipy.optimize.minimize cannot use jax arrays
        def crit_asnumpy(p):
            J = criterion(p).item()
            record_history(p, J)
            return J

        gradient_asnumpy = lambda p: gnp.to_np(gradient(p))

    elif gnp._gpmp_backend_ == "torch":

        def crit_asnumpy(p):
            v = criterion(gnp.asarray(p))
            J = v.detach().item()
            record_history(p, J)
            return J

        def gradient_asnumpy(p):
            g = gradient(gnp.asarray(p))
            if g is None:
                return gnp.zeros(p.shape)
            else:
                return g

    elif gnp._gpmp_backend_ == "numpy":

        def crit_asnumpy(p):
            try:
                J = criterion(p)
            except Exception as e:
                J = np.Inf
            record_history(p, J)
            return J

        gradient_asnumpy = None

    # Set default optimization options and update with user-provided options
    options = {"disp": not silent}
    if method == "L-BFGS-B":
        options.update(
            {
                "maxcor": 20,
                "ftol": 1e-6,
                "gtol": 1e-5,
                "eps": 1e-8,
                "maxfun": 15000,
                "maxiter": 15000,
                "maxls": 40,
                "iprint": -1,
                "maxls": 40,
                "finite_diff_rel_step": None,
            }
        )
    elif method == "SLSQP":
        options.update(
            {"ftol": 1e-6, "eps": 1e-8, "maxiter": 15000, "finite_diff_rel_step": None}
        )
    else:
        raise ValueError("Optimization method not implemented.")

    options.update(method_options)

    # Perform the minimization
    r = minimize(
        crit_asnumpy,
        p0,
        args=(),
        method=method,
        jac=gradient_asnumpy,
        bounds=bounds,
        tol=None,
        callback=None,
        options=options,
    )

    # Ensure that the best values are returned
    if r.fun > best_criterion:
        r.x = best_params
        r.fun = best_criterion
        r.best_value_returned = False
    else:
        r.best_value_returned = True

    # Set additional information about the optimization process
    r.history_params = history_params
    r.history_criterion = history_criterion
    r.initial_params = p0
    r.final_params = r.x
    r.selection_criterion = criterion
    r.total_time = time.time() - tic

    if not silent:
        print("Optimization completed. Best found criterion:", best_criterion)
        print("Gradient")
        print("--------")
        if gradient_asnumpy is not None:
            print(gradient_asnumpy(r.x))
        else:
            print("gradient not available")
        print(".")

    if info:
        return r.x, r
    else:
        return r.x


def select_parameters_with_criterion(
    model,
    criterion,
    xi,
    zi,
    meanparam0=None,
    covparam0=None,
    parameterized_mean=False,
    meanparam_len=1,
    info=False,
    verbosity=0,
):
    """Optimize Gaussian process model parameters using a specified
    selection criterion.

    This function performs parameter optimization for a Gaussian
    process model using the provided selection criterion. It can use a
    provided initial guess for the covariance parameters or employ a
    default initialization strategy. Additional information about the
    optimization can be obtained by setting the `info` parameter to
    True.

    Parameters
    ----------
    model : object
        Instance of a Gaussian process model that needs parameter optimization.
    criterion : function
        The selection criterion function (e.g., ML, REML or REMAP). The function must follow one of two forms:

        - For models without a parameterized mean:
          `criterion(model, covparam, xi, zi)` where `covparam` are the covariance parameters.

        - For models with a parameterized mean:
          `criterion(model, meanparam, covparam, xi, zi)` where both `meanparam` and
          `covparam` are passed.
    xi : ndarray, shape (n, d)
        Locations of the observed data points in the input space.
    zi : ndarray, shape (n,)
        Observed values corresponding to the data points `xi`.
    meanparam0 : ndarray, shape (meanparam_len,), optional
        Initial guess for the mean parameters. Required if `parameterized_mean=True`. Default is None.
    covparam0 : ndarray, shape (covparam_dim,), optional
        Initial guess for the covariance parameters. If not provided, the function
        defaults to the `anisotropic_parameters_initial_guess` method for initialization.
        Default is None.
    parameterized_mean : bool, optional
        Whether the mean is parameterized and included in the selection criterion.
        Default is False.
    meanparam_len : int, optional
        Length of the mean parameter vector if `parameterized_mean` is True. Default is 1.
    info : bool, optional
        Controls the return of additional optimization information. If set to True, the
        function returns an info dictionary with details about the optimization process.
        Default is False.
    verbosity : int, optional, values in {0, 1, 2}
        Sets the verbosity level for the function.
        - 0: No output messages (default).
        - 1: Minimal output messages.
        - 2: Detailed output messages.

    Returns
    -------
    model : object
        The input Gaussian process model object, updated with the optimized parameters.
    info_ret : dict, optional
        A dictionary with additional details about the optimization process. It contains
        fields like initial parameters (`covparam0`), final optimized parameters
        (`covparam`), selection criterion used, and the total time taken for optimization.
        This dictionary is only returned if `info` is set to True.

    Notes
    -----
    The `criterion` function should follow one of the following two forms:

    - For models without a parameterized mean:
      `criterion(model, covparam, xi, zi)` where `covparam` are the covariance parameters.

    - For models with a parameterized mean:
      `criterion(model, meanparam, covparam, xi, zi)` where both `meanparam` and `covparam` are passed.
    """
    tic = time.time()

    if covparam0 is None:
        covparam0 = anisotropic_parameters_initial_guess(model, xi, zi)

    # If the model has a parameterized mean, we need an initial guess for the mean parameters
    if parameterized_mean:
        if meanparam0 is None:
            raise ValueError(
                "meanparam0 must be provided when parameterized_mean is True. Use anisotropic_parameters_initial_guess_constant_mean if needed."
            )
        param0 = gnp.concatenate([meanparam0, covparam0])
    else:
        param0 = covparam0

    # Create the criterion and its gradient using the passed criterion function
    criterion_func, criterion_grad = make_selection_criterion_with_gradient(
        model,
        criterion,
        xi,
        zi,
        parameterized_mean=parameterized_mean,
        meanparam_len=meanparam_len,
    )

    # Optimize parameters using the provided criterion
    silent = True
    if verbosity == 1:
        print("Parameter selection using custom criterion...")
    elif verbosity == 2:
        silent = False

    param_opt, info_ret = autoselect_parameters(
        param0, criterion_func, criterion_grad, silent=silent, info=True
    )

    if verbosity == 1:
        print("done.")

    # Split the optimized parameters into meanparam and covparam if the mean is parameterized
    if parameterized_mean:
        meanparam_opt = param_opt[:meanparam_len]
        covparam_opt = param_opt[meanparam_len:]
        model.meanparam = gnp.asarray(meanparam_opt)
    else:
        meanparam_opt = None
        covparam_opt = param_opt

    model.covparam = gnp.asarray(covparam_opt)

    # NB: info_ret is essentially a dict with attribute accessors
    if info:
        info_ret["meanparam0"] = meanparam0
        info_ret["covparam0"] = covparam0
        info_ret["meanparam"] = meanparam_opt
        info_ret["covparam"] = covparam_opt
        info_ret["selection_criterion"] = criterion_func
        info_ret["time"] = time.time() - tic
        return model, info_ret
    else:
        return model


def update_parameters_with_criterion(
    model, criterion, xi, zi, parameterized_mean=False, meanparam_len=1, info=False
):
    """Update model parameters using a specified selection criterion.

    Parameters
    ----------
    model : object
        Gaussian process model.
    criterion : function
        The selection criterion function (e.g., REML or REMAP).
    xi : ndarray, shape (n, d)
        Locations of the observed data points.
    zi : ndarray, shape (n,)
        Observed values at the data points.
    parameterized_mean : bool, optional
        Whether the mean is parameterized and included in the selection criterion.
        Default is False.
    meanparam_len : int, optional
        Length of the mean parameter vector if `parameterized_mean` is True. Default is 1.
    info : bool, optional
        If True, returns additional information. Default is False.

    Returns
    -------
    model : object
        Updated Gaussian process model object with optimized parameters.
    info_ret : object, optional
        Additional information about the optimization (if info=True).
    """

    return select_parameters_with_criterion(
        model,
        criterion,
        xi,
        zi,
        meanparam0=model.meanparam if parameterized_mean else None,
        covparam0=model.covparam,
        parameterized_mean=parameterized_mean,
        meanparam_len=meanparam_len,
        info=info,
    )


def negative_log_likelihood_zero_mean(model, covparam, xi, zi):
    """Wrapper to core.model.negative_log_likelihood_zero_mean."""
    return model.negative_log_likelihood_zero_mean(covparam, xi, zi)


def negative_log_likelihood(model, meanparam, covparam, xi, zi):
    """Wrapper to core.model.negative_log_likelihood."""
    return model.negative_log_likelihood(meanparam, covparam, xi, zi)


def negative_log_restricted_likelihood(model, covparam, xi, zi):
    """Wrapper to core.model.negative_log_restricted_likelihood."""
    return model.negative_log_restricted_likelihood(covparam, xi, zi)


def select_parameters_with_reml(model, xi, zi, covparam0=None, info=False, verbosity=0):
    """Optimize Gaussian process model parameters using Restricted Maximum Likelihood (REML).

    See select_parameters_with_criterion()
    """
    return select_parameters_with_criterion(
        model,
        negative_log_restricted_likelihood,
        xi,
        zi,
        covparam0=covparam0,
        info=info,
        verbosity=verbosity,
    )


def update_parameters_with_reml(model, xi, zi, info=False):
    """Update model parameters using Restricted Maximum Likelihood (REML).

    See update_parameters_with_criterion
    """
    return update_parameters_with_criterion(
        model, model.negative_log_restricted_likelihood, xi, zi, info=info
    )


def log_prior_jeffrey_variance(covparam, lbda=1.0):
    """Compute a log prior using Jeffrey's prior on the variance parameter.

    This function assumes a Jeffrey's prior on the variance parameter:

    .. math::
        \pi(\sigma^2) \propto \left(\frac{1}{\sigma^2}\right)^\lambda

    Parameters
    ----------
    covparam : ndarray
        Parameter array, where the first element is the
        log(variance) (i.e., log(sigma^2)), and the remaining elements
        are the log of the length-scale parameters (log(1/rho)).
    lbda : float, optional
        Scaling factor for the log prior. Default is 1.0.

    Returns
    -------
    log_p : float
        The logarithm of the prior probability.

    """
    log_sigma2 = covparam[0]

    # Jeffrey's prior: p(sigma^2) \propto (1 / sigma^2)^lbda => log p(sigma^2) = - lbda * log(sigma^2)
    log_prior_sigma2 = - lbda * log_sigma2

    return log_prior_sigma2


def neg_log_restricted_posterior_with_jeffreys_prior(model, covparam, xi, zi):
    """Compute the negative log restricted posterior using Jeffrey's
    prior on the variance parameter.

    This function combines the negative log restricted likelihood with
    a Jeffrey's prior on the variance parameter:

    .. math::
        \pi(\sigma^2) \propto \frac{1}{\sigma^2}

    Parameters
    ----------
    model : object
        Gaussian process model object.
    covparam : ndarray
        The covariance parameter array, where the first element is the
        log(variance) (i.e., log(sigma^2)), and the remaining elements
        are the log of the length-scale parameters (log(1/rho)).
    xi : ndarray
        The input data points.
    zi : ndarray
        The observed values at the input data points.

    Returns
    -------
    float
        The value of the negative log restricted posterior.

    """
    # Compute the negative log restricted likelihood
    nlrl = model.negative_log_restricted_likelihood(covparam, xi, zi)

    # Compute the log prior using Jeffrey's prior on the variance
    log_prior = log_prior_jeffrey_variance(covparam)

    # Posterior
    neg_log_posterior = nlrl - log_prior

    return neg_log_posterior


def select_parameters_with_remap(
    model, xi, zi, covparam0=None, info=False, verbosity=0
):
    """Optimize Gaussian process model parameters using Restricted Maximum A Posteriori (REMAP)."""
    return select_parameters_with_criterion(
        model,
        neg_log_restricted_posterior_with_jeffreys_prior,
        xi,
        zi,
        covparam0=covparam0,
        info=info,
        verbosity=verbosity,
    )


def update_parameters_with_remap(model, xi, zi, info=False):
    """Update model parameters using Restricted Maximum A Posteriori (REMAP)."""
    return update_parameters_with_criterion(
        model, neg_log_restricted_posterior_with_jeffreys_prior, xi, zi, info=info
    )



# --------------------------------------------------------------
# Author: Emmanuel Vazquez <emmanuel.vazquez@centralesupelec.fr>
# Copyright (c) 2022-2024, CentraleSupelec
# License: GPLv3 (see LICENSE)
# --------------------------------------------------------------
import warnings
import gpmp.num as gnp


class Model:
    """Gaussian Process (GP) Model Class.

    This class implements a Gaussian Process (GP) model for function
    approximation.

    Attributes
    ----------
    mean : callable or None
        A function defining the mean of the Gaussian Process (GP),
        used when `self.meantype` is either "parameterized" or
        "linear_predictor". When `self.meantype` is "zero", this
        attribute is not used and should be set to `None`. The mean
        function is called as

        P = self.mean(x, meanparam),

        where `x` is an (n x d) array representing n data points in a
        d-dimensional space, and `meanparam` is an array of parameters
        for the mean function.
        The function returns a (n x q) matrix, where:

        - If `self.meantype` is "parameterized", `q` is 1, and the matrix
          represents the parameterized mean values at the points `x`.
        - If `self.meantype` is "linear_predictor", `q` is greater than or
          equal to 1, indicating the model uses a linear combination
          of q basis functions with linear_predictor coefficients. Each
          column of `P` corresponds to a different basis function
          evaluated at `x`. These basis functions are typically
          monomials, represented as :math:`x \\mapsto 1, x \\mapsto
          x, x \\mapsto x^2, \\ldots`.

    covariance : callable
        Returns the covariance of the GP. The function is called as

        K = self.covariance(x, y, self.covparam, pairwise),

        where x and y are (n x d) and (m x d) arrays of data points,
        and pairwise indicates if an (n x m) covariance matrix
        (pairwise == False) or an (n x 1) vector (n == m, pairwise =
        True) should be returned

    meanparam : array_like, optional
        Parameters for the mean function, given as a one-dimensional
        array. These parameters define the specific form and behavior
        of the mean function used in the GP model.

    covparam : array_like, optional
        Parameters for the covariance function, given as a
        one-dimensional array. These parameters determine the
        characteristics of the covariance function, influencing
        aspects like length scale, variance, and smoothness of the GP.

    meantype : str, optional
        The type of mean used in the model. It can be:

        - 'zero': A zero mean function, implying the GP has a zero prior
          mean function. Then, self.mean is never called and should be set to None.
        - 'parameterized': A parameterized mean function with parameterized parameters. Useful
          when there's prior knowledge about the mean behavior of the
          function being modeled.
        - 'linear_predictor': A linearly parameterized mean function with
          linear_predictor parameters, suitable for situations where the mean
          structure is to be learned from data.

    Methods
    -------
    kriging_predictor_with_zero_mean(xi, xt, return_type=0)
        Compute the kriging predictor assuming a zero mean
        function. Useful for models where the mean is assumed to be
        negligible.

    kriging_predictor(xi, xt, return_type=0)
        Compute the kriging predictor considering a non-zero mean
        function. This method is essential in practical applications.

    predict(xi, zi, xt, return_lambdas=False, zero_neg_variances=True, convert_in=True, convert_out=True)
        Performs prediction at target points `xt` given the data `(xi,
        zi)`. The treatment of the mean function is based on the
        `meantype` attribute.

    loo(xi, zi, convert_in=True, convert_out=False)
        Compute the leave-one-out (LOO) prediction error. This method
        is valuable for model validation and hyperparameter tuning.

    negative_log_likelihood_zero_mean(covparam, xi, zi)
        Computes the negative log-likelihood for a zero-mean GP model.

    negative_log_likelihood(meanparam, covparam, xi, zi)
        Computes the negative log-likelihood for a GP model with a given mean.

    negative_log_restricted_likelihood(covparam, xi, zi)
        Computes the negative log-restricted likelihood (REML)

    norm_k_sqrd_with_zero_mean(xi, zi, covparam)
        Computes the squared norm of the residual vector for a
        zero-mean GP model.

    norm_k_sqrd(xi, zi, covparam)
        Computes the squared norm of the residual vector after
        applying a contrast matrix.

    k_inverses(xi, zi, covparam)
        Calculates various quantities involving the inverse of the
        covariance matrix K.

    sample_paths(xt, nb_paths, method="chol", check_result=True)
        Generates sample paths from the GP model at specified input points.

    conditional_sample_paths(ztsim, xi_ind, zi, xt_ind, lambda_t)
        Generates conditional sample paths from unconditional sample
        paths using kriging weights.

    Examples
    --------
    >>> import gpmp as gp
    >>> import gpmp.num as gnp
    >>> mean = lambda x, meanparam: (meanparam[0] + meanparam[1] * x)
    >>> def covariance(x, y, covparam, pairwise=False):
    >>>     p = 0
    >>>     return gp.kernel.maternp_covariance(x, y, p, covparam, pairwise)
    >>> model = gp.core.Model(mean, covariance, meanparam=[0.5, 0.2], covparam=[1.0, 0.1])
    >>> xi = gnp.array([0.0, 1.0, 2.0, 3.0, 5.0]).reshape(-1, 1)
    >>> zi = gnp.array([0.0, 1.2, 2.5, 4.2, 4.3])
    >>> xt = gnp.array([0.0, 0.5, 1.0, 1.5, 2.0, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0]).reshape(-1, 1)
    >>> zt_mean, zt_var = model.predict(xi, zi, xt)
    """

    def __init__(
        self,
        mean,
        covariance,
        meanparam=None,
        covparam=None,
        meantype="linear_predictor",
    ):
        """
        Parameters
        ----------
        mean : callable
            A function that returns the mean of the Gaussian Process (GP).
        covariance : callable
            A function that returns the covariance of the Gaussian Process (GP).
        meanparam : array_like, optional
            The parameters for the mean function, specified as a one-dimensional array of values.
        covparam : array_like, optional
            The parameters for the covariance function, specified as a one-dimensional array of values.
        meantype : str, optional
            Type of mean used in the model. It can be:
            'zero' - Zero mean function.
            'parameterized' - Known mean function with parameterized parameters.
            'linear_predictor' - Linearly parameterized mean function with linear_predictor parameters.
        """
        self.meantype = meantype

        if meantype not in ["zero", "parameterized", "linear_predictor"]:
            raise ValueError(
                "meantype must be one of 'zero', 'parameterized', or 'linear_predictor'"
            )

        if meantype == "zero" and mean is not None:
            raise ValueError("For meantype 'zero', mean must be None")

        if meantype in ["parameterized", "linear_predictor"] and not callable(mean):
            raise TypeError(
                "For meantype 'parameterized' or 'linear_predictor', mean must be a callable function"
            )

        self.meanparam = meanparam
        self.covparam = covparam
        self.mean = mean
        self.covariance = covariance

    def __repr__(self):
        output = str("<gpmp.core.Model object> " + hex(id(self)))
        return output

    def __str__(self):
        output = str("<gpmp.core.Model object>")
        return output

    def kriging_predictor_with_zero_mean(self, xi, xt, return_type=0):
        """Compute the kriging predictor with zero mean."""
        Kii = self.covariance(xi, xi, self.covparam)
        Kit = self.covariance(xi, xt, self.covparam)

        lambda_t, _ = gnp.cholesky_solve(Kii, Kit)

        if return_type == -1:
            zt_posterior_variance = None
        elif return_type == 0:
            zt_prior_variance = self.covariance(xt, None, self.covparam, pairwise=True)
            zt_posterior_variance = zt_prior_variance - gnp.einsum(
                "i..., i...", lambda_t, Kit
            )
        elif return_type == 1:
            zt_prior_variance = self.covariance(xt, None, self.covparam, pairwise=False)
            zt_posterior_variance = zt_prior_variance - gnp.matmul(lambda_t.T, Kit)

        return lambda_t, zt_posterior_variance

    def kriging_predictor(self, xi, xt, return_type=0):
        """Compute the kriging predictor with non-zero mean.

        Parameters
        ----------
        xi : ndarray(ni, d)
            Observation points
        xt : ndarray(nt, d)
            Prediction points
        return_type : -1, 0 or 1
            If -1, returned posterior variance is None. If 0
            (default), return the posterior variance at points xt.
            If 1, return the posterior covariance.

        Notes
        -----
        If return_type==1, then the covariance function k must be
        built so that k(xi, xi, covparam) returns the covariance
        matrix of observations, and k(xt, xt, covparam) returns the
        covariance matrix of the predictands. This means that the
        information of which are the observation points and which are
        the prediction points must be coded in xi / xt
        """
        # LHS
        Kii = self.covariance(xi, xi, self.covparam)
        Pi = self.mean(xi, self.meanparam)
        (ni, q) = Pi.shape
        # build [ [K P] ; [P' 0] ]
        LHS = gnp.vstack((gnp.hstack((Kii, Pi)), gnp.hstack((Pi.T, gnp.zeros((q, q))))))

        # RHS
        Kit = self.covariance(xi, xt, self.covparam)
        Pt = self.mean(xt, self.meanparam)
        RHS = gnp.vstack((Kit, Pt.T))

        # lambdamu_t = RHS^(-1) LHS
        lambdamu_t = gnp.solve(
            LHS, RHS, overwrite_a=True, overwrite_b=True, assume_a="sym"
        )

        lambda_t = lambdamu_t[0:ni, :]

        # posterior variance
        if return_type == -1:
            zt_posterior_variance = None
        elif return_type == 0:
            zt_prior_variance = self.covariance(xt, None, self.covparam, pairwise=True)
            zt_posterior_variance = zt_prior_variance - gnp.einsum(
                "i..., i...", lambdamu_t, RHS
            )
        elif return_type == 1:
            zt_prior_variance = self.covariance(xt, None, self.covparam, pairwise=False)
            zt_posterior_variance = zt_prior_variance - gnp.matmul(lambdamu_t.T, RHS)

        return lambda_t, zt_posterior_variance

    def predict(
        self,
        xi,
        zi,
        xt,
        return_lambdas=False,
        zero_neg_variances=True,
        convert_in=True,
        convert_out=True,
    ):
        """Performs a prediction at target points xt given the data (xi, zi).

        Parameters
        ----------
        xi : ndarray or gnp.array of shape (ni, dim)
            Observation points.
        zi : ndarray or gnp.array of shape (ni,) or (ni, 1)
            Observed values.
        xt : ndarray or gnp.array of shape (nt, dim)
            Prediction points.
        return_lambdas : bool, optional
            Set return_lambdas=True if lambdas should be returned, by default False.
        zero_neg_variances : bool, optional
            Whether to zero negative posterior variances (due to numerical errors), default=True.
        convert_in : bool, optional
            Whether to convert input arrays to _gpmp_backend_ type or keep as-is.
        convert_out : bool, optional
            Whether to return numpy arrays or keep _gpmp_backend_ types.

        Returns
        -------
        z_posterior_mean : gnp.array or ndarray
            1D array of shape nt representing the posterior mean.
        z_posterior_variance : gnp.array or ndarray
            1D array of shape nt representing the posterior variance.
        lambda_t : gnp.array(ni, nt), optional
            2D array of kriging weights, only returned if return_lambdas=True.

        Notes
        -----
        From a Bayesian point of view, the outputs are respectively
        the posterior mean and variance of the Gaussian process given
        the data (xi, zi).

        Treatment of the mean based on 'meantype':
        1. "zero": The function uses the kriging predictor with zero mean.
        2. "linear_predictor": Uses the general / intrinsic kriging predictor.
        3. "parameterized": The zero-mean kriging predictor is used after
        centering zi around the parameterized mean. The mean is then added
        back to the posterior mean prediction. 'meanparam' should
        be provided for this type.

        Ensure to set the appropriate 'meantype' for the desired
        behavior. Supported types are 'zero', 'parameterized', and 'linear_predictor'.
        """
        xi_, zi_, xt_ = Model.ensure_shapes_and_type(
            xi=xi, zi=zi, xt=xt, convert=convert_in
        )

        # Decide which kriging predictor to use and if we need to adjust for mean
        zt_prior_mean_ = 0.0
        if self.meantype == "zero":
            lambda_t, zt_posterior_variance_ = self.kriging_predictor_with_zero_mean(
                xi_, xt_
            )
        elif self.meantype == "linear_predictor":
            lambda_t, zt_posterior_variance_ = self.kriging_predictor(xi_, xt_)
        elif self.meantype == "parameterized":
            lambda_t, zt_posterior_variance_ = self.kriging_predictor_with_zero_mean(
                xi_, xt_
            )

            if self.meanparam is None:
                raise ValueError(
                    "For meantype 'parameterized', meanparam *should not* be None."
                )
            zi_prior_mean_ = self.mean(xi_, self.meanparam).reshape(-1)
            zi_ = zi_ - zi_prior_mean_
            zt_prior_mean_ = self.mean(xt_, self.meanparam).reshape(-1)
        else:
            raise ValueError(
                f"Invalid mean_type {self.mean_type}. Supported types are 'zero', 'parameterized', and 'linear_predictor'."
            )

        if gnp.any(zt_posterior_variance_ < 0.0):
            warnings.warn(
                "In predict: negative variances detected. Consider using jitter.",
                RuntimeWarning,
            )
        if zero_neg_variances:
            zt_posterior_variance_ = gnp.maximum(zt_posterior_variance_, 0.0)

        # posterior mean
        zt_posterior_mean_ = gnp.einsum("i..., i...", lambda_t, zi_) + zt_prior_mean_

        # outputs
        if convert_out:
            zt_posterior_mean = gnp.to_np(zt_posterior_mean_)
            zt_posterior_variance = gnp.to_np(zt_posterior_variance_)
        else:
            zt_posterior_mean = zt_posterior_mean_
            zt_posterior_variance = zt_posterior_variance_

        if not return_lambdas:
            return (zt_posterior_mean, zt_posterior_variance)
        else:
            return (zt_posterior_mean, zt_posterior_variance, lambda_t)

    def loo(self, xi, zi, convert_in=True, convert_out=False):
        """Compute the leave-one-out (LOO) prediction error.

        This method computes the LOO prediction error using the
        "virtual cross-validation" formula, which allows for efficient
        computation of LOO predictions without re-fitting the model.

        Parameters
        ----------
        xi : array_like, shape (n, d)
            Input data points used for fitting the GP model, where n
            is the number of points and d is the dimensionality.
        zi : array_like, shape (n, ) or (n, 1)
            Output (response) values corresponding to the input data points xi.
        convert_in : bool, optional
            Whether to convert input arrays to _gpmp_backend_ type or keep as-is.
        convert_out : bool, optional
            Whether to return numpy arrays or keep _gpmp_backend_ types.

        Returns
        -------
        zloo : array_like, shape (n, )
            Leave-one-out predictions for each data point in xi.
        sigma2loo : array_like, shape (n, )
            Variance of the leave-one-out predictions.
        eloo : array_like, shape (n, )
            Leave-one-out prediction errors for each data point in xi.

        Examples
        --------
        >>> xi = np.array([[1, 2], [3, 4], [5, 6]])
        >>> zi = np.array([1.2, 2.5, 4.2])
        >>> model = Model(mean, covariance, meanparam=[0.5, 0.2], covparam=[1.0, 0.1])
        >>> zloo, sigma2loo, eloo = model.loo(xi, zi)
        """
        xi_, zi_, _ = Model.ensure_shapes_and_type(xi=xi, zi=zi, convert=convert_in)

        if self.meantype == "zero":
            zloo, sigma2loo, eloo = self._loo_with_zero_mean(self.covparam, xi_, zi_)
        elif self.meantype == "parameterized":
            zloo, sigma2loo, eloo = self._loo_with_parameterized_mean(
                self.meanparam, self.covparam, xi_, zi_
            )
        elif self.meantype == "linear_predictor":
            zloo, sigma2loo, eloo = self._loo_with_linear_predictor_mean(
                self.meanparam, self.covparam, xi_, zi_
            )
        else:
            raise ValueError(f"Unknown mean type: {self.meantype}")

        if convert_out:
            zloo = gnp.to_np(zloo)
            sigma2loo = gnp.to_np(sigma2loo)
            eloo = gnp.to_np(eloo)

        return zloo, sigma2loo, eloo

    def _loo_with_zero_mean(self, covparam, xi, zi):
        """Compute LOO prediction error for zero mean."""
        K = self.covariance(xi, xi, covparam)  # shape (n, n)

        # Use the "virtual cross-validation" formula
        Kinv = gnp.cholesky_inv(K)

        # e_loo,i  = 1 / Kinv_i,i ( Kinv  z )_i
        Kinvzi = gnp.matmul(Kinv, zi)  # shape (n, )
        Kinvdiag = gnp.diag(Kinv)  # shape (n, )
        eloo = Kinvzi / Kinvdiag  # shape (n, )

        # sigma2_loo,i = 1 / Kinv_i,i
        sigma2loo = 1.0 / Kinvdiag  # shape (n, )

        # zloo_i = z_i - e_loo,i
        zloo = zi - eloo  # shape (n, )

        return zloo, sigma2loo, eloo

    def _loo_with_parameterized_mean(self, meanparam, covparam, xi, zi):
        """Compute LOO prediction error for parameterized mean."""
        zi_prior_mean = self.mean(xi, meanparam).reshape(-1)
        centered_zi = zi - zi_prior_mean
        zloo_centered, sigma2loo, eloo_centered = self._loo_with_zero_mean(
            covparam, xi, centered_zi
        )
        zloo = zloo_centered + zi_prior_mean
        return zloo, sigma2loo, eloo_centered

    def _loo_with_linear_predictor_mean(self, meanparam, covparam, xi, zi):
        """Compute LOO prediction error for linear_predictor mean."""
        K = self.covariance(xi, xi, covparam)
        P = self.mean(xi, meanparam)

        # Use the "virtual cross-validation" formula
        # Qinv = K^-1 - K^-1P (Pt K^-1 P)^-1 Pt K^-1
        Kinv = gnp.inv(K)
        KinvP = gnp.matmul(Kinv, P)
        PtKinvP = gnp.einsum("ki, kj->ij", P, KinvP)
        R = gnp.solve(PtKinvP, KinvP.T)
        Qinv = Kinv - gnp.matmul(KinvP, R)

        # e_loo,i  = 1 / Q_i,i ( Qinv  z )_i
        Qinvzi = gnp.matmul(Qinv, zi)  # shape (n, )
        Qinvdiag = gnp.diag(Qinv)  # shape (n, )
        eloo = Qinvzi / Qinvdiag  # shape (n, )

        # sigma2_loo,i = 1 / Qinv_i,i
        sigma2loo = 1.0 / Qinvdiag  # shape (n, )

        # z_loo
        zloo = zi - eloo  # shape (n, )

        return zloo, sigma2loo, eloo

    def negative_log_likelihood_zero_mean(self, covparam, xi, zi):
        """Computes the negative log-likelihood of the Gaussian process model with zero
        mean.

        This function is specific to the zero-mean case, and the negative log-likelihood
        is computed based on the provided covariance function and parameters.

        Parameters
        ----------
        covparam : gnp.array
            Parameters for the covariance function. This array contains the hyperparameters
            required by the chosen covariance function.
        xi : ndarray(ni,d)
            Locations of the data points in the input space, where ni is the number of data points
            and d is the dimensionality of the input space.
        zi : ndarray(ni, )
            Observed values corresponding to each data point in xi.

        Returns
        -------
        nll : scalar
            Negative log-likelihood of the observed data given the model and covariance parameters.
        """
        K = self.covariance(xi, xi, covparam)
        n = K.shape[0]

        try:
            Kinv_zi, C = gnp.cholesky_solve(K, zi)
        except RuntimeError:
            if gnp._gpmp_backend_ == "jax" or gnp._gpmp_backend_ == "numpy":
                return gnp.inf
            elif gnp._gpmp_backend_ == "torch":
                inf_tensor = gnp.tensor(float("inf"), requires_grad=True)
                return inf_tensor  # returns inf with None gradient

        norm2 = gnp.einsum("i..., i...", zi, Kinv_zi)
        ldetK = 2.0 * gnp.sum(gnp.log(gnp.diag(C)))

        L = 0.5 * (n * gnp.log(2.0 * gnp.pi) + ldetK + norm2)

        return L.reshape(())

    def negative_log_likelihood(self, meanparam, covparam, xi, zi):
        """Computes the negative log-likelihood of the Gaussian process model with a
        given mean.

        This function computes the negative log-likelihood based on
        the provided mean function, covariance function, and their
        parameters.

        Parameters
        ----------
        meanparam : gnp.array
            Parameters for the mean function. This array contains the
            hyperparameters required by the chosen mean function.
        covparam : gnp.array
            Parameters for the covariance function. This array
            contains the hyperparameters required by the chosen
            covariance function.
        xi : ndarray(ni,d)
            Locations of the data points in the input space, where ni
            is the number of data points and d is the dimensionality
            of the input space.
        zi : ndarray(ni, )
            Observed values corresponding to each data point in xi.

        Returns
        -------
        nll : scalar
            Negative log-likelihood of the observed data given the
            model, mean, and covariance parameters.
        """
        zi_prior_mean = self.mean(xi, meanparam).reshape(-1)
        centered_zi = zi - zi_prior_mean

        # Call the zero mean version with the centered observations
        return self.negative_log_likelihood_zero_mean(covparam, xi, centered_zi)

    def negative_log_restricted_likelihood(self, covparam, xi, zi):
        """Compute the negative log-restricted likelihood of the GP model.

        This method calculates the negative log-restricted likelihood,
        which is used for parameter estimation in the Gaussian Process
        model with a mean of type "linear predictor"

        Parameters
        ----------
        covparam : gnp.array
            Covariance parameters for the Gaussian Process.
        xi : array_like, shape (n, d)
            Input data points used for fitting the GP model, where n
            is the number of points and d is the dimensionality.
        zi : array_like, shape (n, )
            Output (response) values corresponding to the input data points xi.

        Returns
        -------
        L : float
            Negative log-restricted likelihood value.

        Examples
        --------
        >>> xi = np.array([[1, 2], [3, 4], [5, 6]])
        >>> zi = np.array([1.2, 2.5, 4.2])
        >>> model = Model(mean, covariance, meanparam=[0.5, 0.2], covparam=[1.0, 0.1])
        >>> covparam = np.array([1.0, 0.1])
        >>> L = model.negative_log_restricted_likelihood(covparam, xi, zi)
        """
        K = self.covariance(xi, xi, covparam)
        P = self.mean(xi, self.meanparam)
        n, q = P.shape

        # Compute a matrix of contrasts
        [Q, R] = gnp.qr(P, "complete")
        W = Q[:, q:n]

        # Contrasts (n-q) x 1
        Wzi = gnp.matmul(W.T, zi)

        # Compute G = W' * (K * W), the covariance matrix of contrasts
        G = gnp.matmul(W.T, gnp.matmul(K, W))

        # Cholesky factorization: G = U' * U, with upper-triangular U
        # Compute G^(-1) * (W' zi)
        try:
            WKWinv_Wzi, C = gnp.cholesky_solve(G, Wzi)
        except RuntimeError:
            if gnp._gpmp_backend_ == "jax" or gnp._gpmp_backend_ == "numpy":
                return gnp.inf
            elif gnp._gpmp_backend_ == "torch":
                # Use LinAlgError instead of raising RuntimeError for linalg operations
                # https://github.com/pytorch/pytorch/issues/64785
                # https://stackoverflow.com/questions/242485/starting-python-debugger-automatically-on-error
                # extype, value, tb = __import__("sys").exc_info()
                # __import__("traceback").print_exc()
                # __import__("pdb").post_mortem(tb)
                inf_tensor = gnp.tensor(float("inf"), requires_grad=True)
                return inf_tensor  # returns inf with None gradient

        # Compute norm2 = (W' zi)' * G^(-1) * (W' zi)
        norm2 = gnp.einsum("i..., i...", Wzi, WKWinv_Wzi)

        # Compute log(det(G)) using the Cholesky factorization
        ldetWKW = 2.0 * gnp.sum(gnp.log(gnp.diag(C)))

        L = 0.5 * ((n - q) * gnp.log(2.0 * gnp.pi) + ldetWKW + norm2)

        return L.reshape(())

    def norm_k_sqrd_with_zero_mean(self, xi, zi, covparam):
        """Compute the squared norm of the residual vector with zero mean.

        This method calculates the squared norm of the residual vector
        (zi - mean(xi)) using the inverse of the covariance matrix K.

        Parameters
        ----------
        xi : array_like, shape (n, d)
            Input data points used for fitting the GP model, where n
            is the number of points and d is the dimensionality.
        zi : array_like, shape (n, )
            Output (response) values corresponding to the input data points xi.
        covparam : array_like
            Covariance parameters for the Gaussian Process.

        Returns
        -------
        norm_sqrd : float
            Squared norm of the residual vector.

        Examples
        --------
        >>> xi = np.array([[1, 2], [3, 4], [5, 6]])
        >>> zi = np.array([1.2, 2.5, 4.2])
        >>> model = Model(mean, covariance, meanparam=[0.5, 0.2], covparam=[1.0, 0.1])
        >>> covparam = np.array([1.0, 0.1])
        >>> norm_sqrd = model.norm_k_sqrd_with_zero_mean(xi, zi, covparam)
        """
        K = self.covariance(xi, xi, covparam)
        Kinv_zi, _ = gnp.cholesky_solve(K, zi)
        norm_sqrd = gnp.einsum("i..., i...", zi, Kinv_zi)

        return norm_sqrd

    def k_inverses(self, xi, zi, covparam):
        """Compute various quantities involving the inverse of K.

        Specifically, this method calculates:
        - z^T K^-1 z
        - K^-1 1
        - K^-1 z

        Parameters
        ----------
        xi : array_like, shape (n, d)
            Input data points used for fitting the GP model.
        zi : array_like, shape (n, 1)
            Output (response) values corresponding to the input data points xi.
        covparam : array_like
            Covariance parameters for the Gaussian process.

        Returns
        -------
        zTKinvz : float
            z^T K^-1 z
        Kinv1 : array_like, shape (n, 1)
            K^-1 1
        Kinvz : array_like, shape (n, 1)
            K^-1 z

        Examples
        --------
        >>> xi = np.array([[1, 2], [3, 4], [5, 6]])
        >>> zi = np.array([[1.2], [2.5], [4.2]])
        >>> model = Model(mean, covariance, meanparam=[0.5, 0.2], covparam=[1.0, 0.1])
        >>> covparam = np.array([1.0, 0.1])
        >>> zTKinvz, Kinv1, Kinvz = model.compute_k_inverses(xi, zi, covparam)
        """
        K = self.covariance(xi, xi, covparam)
        ones_vector = gnp.ones(zi.shape)

        Kinv = gnp.cholesky_inv(K)
        Kinv_zi = gnp.einsum("...i, i... ", Kinv, zi)
        Kinv_1 = gnp.einsum("...i, i... ", Kinv, ones_vector)

        zTKinvz = gnp.einsum("i..., i...", zi, Kinv_zi)

        return zTKinvz, Kinv_1, Kinv_zi

    def norm_k_sqrd(self, xi, zi, covparam):
        """Compute the squared norm of the residual vector after applying the contrast
        matrix W.

        This method calculates the squared norm of the residual vector
        (Wz) using the inverse of the covariance matrix (WKW), where W
        is a matrix of contrasts.

        Parameters
        ----------
        xi : ndarray, shape (ni, d)
            Input data points used for fitting the GP model, where ni
            is the number of points and d is the dimensionality.
        zi : ndarray, shape (ni, 1) or (ni, )
            Output (response) values corresponding to the input data points xi.
        covparam : array_like
            Covariance parameters for the Gaussian Process.

        Returns
        -------
        float
            The squared norm of the residual vector after applying the
            contrast matrix W: (Wz)' (WKW)^-1 Wz.
        """
        K = self.covariance(xi, xi, covparam)
        P = self.mean(xi, self.meanparam)
        n, q = P.shape

        # Compute a matrix of contrasts
        [Q, R] = gnp.qr(P, "complete")
        W = Q[:, q:n]

        # Contrasts (n-q) x 1
        Wzi = gnp.matmul(W.T, zi)

        # Compute G = W' * (K * W), the covariance matrix of contrasts
        G = gnp.matmul(W.T, gnp.matmul(K, W))

        # Compute G^(-1) * (W' zi)
        WKWinv_Wzi, _ = gnp.cholesky_solve(G, Wzi)

        # Compute norm_2 = (W' zi)' * G^(-1) * (W' zi)
        norm_sqrd = gnp.einsum("i..., i...", Wzi, WKWinv_Wzi)

        return norm_sqrd

    def sample_paths(self, xt, nb_paths, method="chol", check_result=True):
        """Generates nb_paths sample paths on xt from the zero-mean GP model GP(0, k),
        where k is the covariance specified by Model.covariance.

        Parameters
        ----------
        xt : ndarray, shape (nt, d)
            Input data points where the sample paths are to be
            generated, where nt is the number of points and d is the
            dimensionality.
        nb_paths : int
            Number of sample paths to generate.
        method : str, optional, default: 'chol'
            Method used for the factorization of the covariance
            matrix. Options are 'chol' for Cholesky decomposition and
            'svd' for singular value decomposition.
        check_result : bool, optional, default: True
            If True, checks if the Cholesky factorization is successful.

        Returns
        -------
        ndarray, shape (nt, nb_paths)
            Array containing the generated sample paths at the input points xt.

        Examples
        --------
        >>> xt = np.array([[1, 2], [3, 4], [5, 6]])
        >>> model = Model(mean, covariance, meanparam=[0.5, 0.2], covparam=[1.0, 0.1])
        >>> nb_paths = 10
        >>> sample_paths = model.sample_paths(xt, nb_paths)
        """
        xt_ = gnp.asarray(xt)

        K = self.covariance(xt_, xt_, self.covparam)

        # Factorization of the covariance matrix
        if method == "chol":
            C = gnp.cholesky(K)
            if check_result:
                if gnp.isnan(C).any():
                    raise AssertionError(
                        "In sample_paths: Cholesky factorization failed. Consider using jitter or the sdv switch."
                    )
        elif method == "svd":
            u, s, vt = gnp.svd(K, full_matrices=True, hermitian=True)
            C = gnp.matmul(u * gnp.sqrt(s), vt)

        # Generates samplepaths
        zsim = gnp.matmul(C, gnp.randn(K.shape[0], nb_paths))

        return zsim

    def conditional_sample_paths(self, ztsim, xi_ind, zi, xt_ind, lambda_t, convert_out=True):
        """Generates conditional sample paths on xt from unconditional sample paths
        ztsim, using the matrix of kriging weights lambda_t, which is provided by
        kriging_predictor() or predict().

        Conditioning is done with respect to ni observations, located
        at the indices given by xi_ind in ztsim, with corresponding
        observed values zi. xt_ind specifies indices in ztsim
        corresponding to conditional simulation points.

        This method assumes the mean function is of type 'zero' or
        'linear_predictor'.

        NOTE: the function implements "conditioning by kriging" (see,
        e.g., Chiles and Delfiner, Geostatistics: Modeling Spatial
        Uncertainty, Wiley, 1999).

        Parameters
        ----------
        ztsim : ndarray, shape (n, nb_paths)
            Unconditional sample paths.
        zi : ndarray, shape (ni, 1) or (ni, )
            Observed values corresponding to the input data points xi.
        xi_ind : ndarray, shape (ni, 1), dtype=int
            Indices of observed data points in ztsim.
        xt_ind : ndarray, shape (nt, 1), dtype=int
            Indices of prediction data points in ztsim.
        lambda_t : ndarray, shape (ni, nt)
            Kriging weights.
        convert_out : bool, optional
            Whether to return numpy arrays or keep _gpmp_backend_ types.

        Returns
        -------
        ztsimc : ndarray, shape (nt, nb_paths)
            Conditional sample paths at the prediction data points xt.

        Examples
        --------
        >>> ztsim = np.random.randn(10, 5)
        >>> zi = np.array([[1], [2], [3]])
        >>> xi_ind = np.array([[0], [3], [7]])
        >>> xt_ind = np.array([[1], [2], [4], [5], [6], [8], [9]])
        >>> lambda_t = np.random.randn(3, 7)
        >>> ztsimc = model.conditional_sample_paths(ztsim, xi_ind, zi, xt_ind, lambda_t)
        """
        zi_ = gnp.asarray(zi).reshape(-1, 1)
        ztsim_ = gnp.asarray(ztsim)

        delta = zi_ - ztsim_[xi_ind, :]

        ztsimc = ztsim_[xt_ind, :] + gnp.einsum("ij,ik->jk", lambda_t, delta)

        if convert_out:
            ztsimc = gnp.to_np(ztsimc)

        return ztsimc

    def conditional_sample_paths_parameterized_mean(
            self, ztsim, xi, xi_ind, zi, xt, xt_ind, lambda_t, convert_out=True
    ):
        """Generates conditional sample paths with a parameterized mean function.

        This method accommodates parameterized means, adjusting the
        unconditional sample paths 'ztsim' with respect to observed
        values 'zi' at 'xi' and prediction points 'xt', using kriging
        weights 'lambda_t'.

        Parameters
        ----------
        ztsim : ndarray
            Unconditional sample paths.
        xi : ndarray
            Input data points for observed values.
        xi_ind : ndarray, dtype=int
            Indices of observed data points in ztsim.
        zi : ndarray
            Observed values at the input data points xi.
        xt : ndarray
            Prediction data points.
        xt_ind : ndarray, dtype=int
            Indices of prediction data points in ztsim.
        lambda_t : ndarray
            Kriging weights matrix.
        convert_out : bool, optional
            Whether to return numpy arrays or keep _gpmp_backend_ types.

        Returns
        -------
        ztsimc : ndarray
            Conditional sample paths at prediction points xt, adjusted for a parameterized mean.
        """
        xi_, zi_, xt_ = Model.ensure_shapes_and_type(xi=xi, zi=zi, xt=xt)
        ztsim_ = gnp.asarray(ztsim)

        zi_prior_mean_ = self.mean(xi_, self.meanparam).reshape(-1)
        zi_centered_ = zi_ - zi_prior_mean_

        zt_prior_mean_ = self.mean(xt_, self.meanparam).reshape(-1, 1)

        delta = zi_centered_.reshape(-1, 1) - ztsim_[xi_ind, :]

        ztsimc = (
            ztsim_[xt_ind, :]
            + gnp.einsum("ij,ik->jk", lambda_t, delta)
            + zt_prior_mean_
        )

        if convert_out:
            ztsimc = gnp.to_np(ztsimc)

        return ztsimc

    @staticmethod
    def ensure_shapes_and_type(xi=None, zi=None, xt=None, convert=True):
        """Ensure correct shapes and types for provided arrays.

        Parameters
        ----------
        xi : ndarray or gnp.array(ni, dim), optional
            Observation points.
        zi : ndarray or gnp.array(ni,) or gnp.array(ni, 1), optional
            Observed values.
        xt : ndarray or gnp.array(nt, dim), optional
            Prediction points.
        convert : bool, optional
            Whether to convert input arrays to _gpmp_backend_ type or keep as-is.

        Returns
        -------
        xi, zi, xt : tuple
            Tuples containing arrays with ensured shapes and types.
        """

        if xi is not None:
            assert len(xi.shape) == 2, "xi should be a 2D array"

        if zi is not None:
            if len(zi.shape) == 2:
                assert (
                    zi.shape[1] == 1
                ), "zi should only have one column if it's a 2D array"
                zi = zi.reshape(-1)  # reshapes (ni, 1) to (ni,)
            else:
                assert len(zi.shape) == 1, "zi should either be 1D or a 2D column array"

        if xt is not None:
            assert len(xt.shape) == 2, "xt should be a 2D array"

        if xi is not None and zi is not None:
            assert (
                xi.shape[0] == zi.shape[0]
            ), "Number of rows in xi should be equal to the number of rows in zi"
        if xi is not None and xt is not None:
            assert (
                xi.shape[1] == xt.shape[1]
            ), "xi and xt should have the same number of columns"

        if convert:
            if xi is not None:
                xi = gnp.asarray(xi)
            if zi is not None:
                zi = gnp.asarray(zi)
            if xt is not None:
                xt = gnp.asarray(xt)

        return xi, zi, xt



## --------------------------------------------------------------
# Author: Emmanuel Vazquez <emmanuel.vazquez@centralesupelec.fr>
# Copyright (c) 2022, CentraleSupelec
# License: GPLv3 (see LICENSE)
## --------------------------------------------------------------
import sys
import numpy as np
import scipy.stats as stats
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
from matplotlib import interactive


class Figure:
    """Figures manager class.

    Attributes
    ----------

    Methods
    -------
    """

    def __init__(self, nrows=1, ncols=1, isinteractive=True, boxoff=True, **kargs):
        # Check if we run in interpreter mode
        # https://stackoverflow.com/questions/2356399/tell-if-python-is-in-interactive-mode
        self.interpreter = False
        try:
            if sys.ps1:
                self.interpreter = True
        except AttributeError:
            self.interpreter = False
            if sys.flags.interactive:
                self.interpreter = True

        if isinteractive & self.interpreter:
            interactive(True)

        self.boxoff = boxoff

        self.fig = plt.figure(**kargs)

        self.nrows = nrows
        self.ncols = ncols
        self.axes = []
        for i in range(nrows * ncols):
            self.axes.append(self.fig.add_subplot(nrows, ncols, i + 1))
        self.ax = self.axes[0]
        if self.boxoff:
            self.set_boxoff()

    def set_boxoff(self):
        self.ax.spines["right"].set_visible(False)
        self.ax.spines["top"].set_visible(False)
        self.ax.tick_params(direction="in")

    def subplot(self, i):
        self.ax = self.axes[i - 1]
        if self.boxoff:
            self.set_boxoff()

    def show(self, grid=None, legend=None, legend_fontsize=None, xlim=None):
        if grid:
            self.grid()
        if legend and legend_fontsize is not None:
            self.legend(fontsize=legend_fontsize)
        elif legend:
            self.legend()
        if xlim is not None:
            self.xlim(xlim)
        plt.show()

    def plot(self, x, z, *args, **kargs):
        self.ax.plot(x, z, *args, **kargs)

    def plotdata(self, x, z, label='data'):
        self.ax.plot(x, z, "rs", markerfacecolor="none", markersize=6, label=label)

    def xlabel(self, s):
        self.ax.set_xlabel(s)

    def ylabel(self, s):
        self.ax.set_ylabel(s)

    def xylabels(self, sx="", sy=""):
        self.ax.set_xlabel(sx)
        self.ax.set_ylabel(sy)

    def title(self, s):
        self.ax.set_title(s)

    def legend(self, **kwargs):
        self.ax.legend(**kwargs)

    def grid(
        self,
        visible=True,
        which="major",
        linestyle=(0, (1, 5)),
        linewidth=0.5,
        **kwargs
    ):
        self.ax.grid(visible, which, linestyle=linestyle, linewidth=linewidth, **kwargs)

    def xlim(self, new_limits=None):
        if new_limits is None:
            return self.ax.get_xlim()
        else:
            self.ax.set_xlim(new_limits)
            return new_limits

    def ylim(self, new_limits=None):
        if new_limits is None:
            return self.ax.get_ylim()
        else:
            self.ax.set_ylim(new_limits)
            return new_limits

    def plotgp(
        self,
        x,
        mean,
        variance,
        colorscheme="default",
        rgb_hue=None,
        ax=None,
        fignum=None,
        mean_label="posterior mean",
        show_mean_label=True,
        ci=[0.95, 0.99, 0.999],  # CI levels
        ci_labels=["CI 95%", "CI 99%", "CI 99.9%"],
        show_ci_labels=True,
        **kwargs
    ):
        """Coverage intervals.

        norminv (1 - 0.05/2)  = 1.959964
        norminv (1 - 0.01/2)  = 2.575829
        norminv (1 - 0.001/2) = 3.290527
        """
        if not show_mean_label:
            mean_label = ""
        if not show_ci_labels:
            ci_labels = [""] * 3

        mean = mean.flatten()
        x = x.flatten()

        # Compute delta values using scipy.stats.norm.ppf for the given CI levels
        delta0 = [stats.norm.ppf((1 + level) / 2) for level in ci]

        if colorscheme == "hue":
            hex_code = "#" + "".join([format(i, "02x") for i in rgb_hue])
            mcol = hex_code
            mwidth = 2.0
            delta0 = [delta0[0]]
            ci_labels = [ci_labels[0]]
            fillcol = [hex_code]
            alpha = 0.5
            kwargs["linewidth"] = 0.5
            drawulb = False
        if colorscheme == "bw":
            mcol = "#000000"
            mwidth = 2.0
            edgecol = "#000000"
            delta0 = [delta0[0]]
            ci_labels = [ci_labels[0]]
            fillcol = ["#F2F2F2"]
            alpha = 0.0
            drawulb = True
        if colorscheme == "simple":
            mcol = "#F2404C"
            mwidth = 2.0
            delta0 = [delta0[0]]
            ci_labels = [ci_labels[0]]
            fillcol = ["#BFBFBF"]
            alpha = 0.8
            kwargs["linewidth"] = 0.5
            drawulb = False
        if colorscheme == "default":
            mcol = "#F2404C"
            mwidth = 2.0
            delta0 = delta0[::-1]
            ci_labels = ci_labels[::-1]
            fillcol = ["#F2F2F2", "#D8D8D8", "#BFBFBF"]
            alpha = 0.8
            kwargs["linewidth"] = 0.5
            drawulb = False

        # mean
        self.ax.plot(x, mean, mcol, linewidth=mwidth, label=mean_label)

        for i, delta in enumerate(delta0):
            kwargs["alpha"] = alpha

            lower = mean - delta * np.sqrt(variance.flatten())
            upper = mean + delta * np.sqrt(variance.flatten())

            self.ax.fill(
                np.hstack((x, x[::-1])),
                np.hstack((upper, lower[::-1])),
                color=fillcol[i],
                label=ci_labels[i],
                **kwargs
            )

            if drawulb:
                self.ax.plot(
                    x,
                    upper,
                    color=edgecol,
                    linestyle="dashed",
                    dashes=(10, 8),
                    linewidth=0.5,
                )
                self.ax.plot(
                    x,
                    lower,
                    color=edgecol,
                    linestyle="dashed",
                    dashes=(10, 8),
                    linewidth=0.5,
                )


def crosssections(model, xi, zi, box, ind_i, ind_dim, nt=100):
    """Display "cross-section" predictions at points xi[ind_i] along dimensions
    specified in ind_dim.

    Parameters
    ----------
    model : _type_
        _description_
    xi : _type_
        _description_
    zi : _type_
        _description_
    box : _type_
        _description_
    ind_i : _type_
        _description_
    ind_dim : _type_
        _description_
    nt : int, optional
        _description_, by default 100
    """
    box = np.array(box)
    num_crosssections = len(ind_i)
    num_dims = len(ind_dim)

    fig = Figure(num_dims, num_crosssections)

    for i in range(num_crosssections):
        for d in range(num_dims):
            t = np.sort(
                np.concatenate(
                    (
                        np.linspace(box[0, d], box[1, d], nt - 1),
                        np.array([xi[ind_i[i], ind_dim[d]]]),
                    )
                )
            )
            xt = np.tile(xi[ind_i[i], :], (nt, 1))
            xt[:, ind_dim[d]] = t
            (zpm, zpv) = model.predict(xi, zi, xt)
            zpv = np.maximum(zpv, 0)
            fig.subplot(num_crosssections * d + i + 1)
            fig.plotgp(t, zpm, zpv)
            fig.plot(xi[ind_i[i], ind_dim[d]] * np.array([1, 1]), fig.ax.get_ylim())
            fig.grid()
            if i == 0:
                fig.ax.set_ylabel("z along x_{:d}".format(d + 1))
            if d == 0:
                fig.ax.set_title("cross section {:d}".format(i + 1))


def plot_loo(zi, zloom, zloov):
    """LOO plot.

    Parameters
    ----------
    zi : _type_
        _description_
    zloom : _type_
        _description_
    zloov : _type_
        _description_
    """
    fig = Figure()
    fig.ax.errorbar(zi, zloom, 1.96 * np.sqrt(zloov), fmt="ko", ls="None")
    fig.ax.set_xlabel("true values"), plt.ylabel("predicted")
    fig.ax.set_title("LOO predictions with 95% coverage intervals")
    (xmin, xmax), (ymin, ymax) = fig.ax.get_xlim(), fig.ax.get_ylim()
    xmin = min(xmin, ymin)
    xmax = max(xmax, ymax)
    fig.ax.plot([xmin, xmax], [xmin, xmax], "--")
    fig.ax.grid(True, "major", linestyle=(0, (1, 5)), linewidth=0.5)
    fig.show()



""" Plot the Matern nu = p + 1/2 kernel/covariance functions

Author: Emmanuel Vazquez <emmanuel.vazquez@centralesupelec.fr>
Copyright (c) 2022, CentraleSupelec
License: GPLv3 (see LICENSE)
"""
import gpmp.num as gnp
import gpmp as gp

def main():
    h = gnp.linspace(-2.0, 2.0, 500)

    fig = gp.misc.plotutils.Figure()

    for p in [0, 1, 4]:
        r = gp.kernel.maternp_kernel(p, gnp.abs(h))
        fig.plot(h, r, label='p={} / nu={}/2'.format(p, 2*p+1))

    fig.title('Matern covariances')
    fig.xlabel('h')
    fig.ylabel('$k_{p+1/2}(h)$')
    fig.legend()
    fig.show(grid=True)


if __name__ == '__main__':
    main()



"""
GP interpolation in 1D, with noiseless data

This example demonstrates how to compute GP interpolation with unknown mean
(aka ordinary / intrinsic kriging) on a one-dimensional noiseless dataset.

A Mat'ern covariance function is used for the Gaussian Process (GP) prior.
The parameters of this covariance function are assumed to be known
(i.e., no parameter estimation is performed here).

The kriging predictor / posterior mean of the GP, interpolates the data.

----
Author: Emmanuel Vazquez <emmanuel.vazquez@centralesupelec.fr>
Copyright (c) 2022-2023, CentraleSupelec
License: GPLv3 (see LICENSE)
----
This example is based on the file stk_example_kb01.m from the STK at
https://github.com/stk-kriging/stk/
by Julien Bect and Emmanuel Vazquez, released under the GPLv3 license.

Original copyright notice:

   Copyright (c) 2015, 2016, 2018 CentraleSupelec
   Copyright (c) 2011-2014 SUPELEC
"""

import math
import numpy as np
import gpmp.num as gnp
import gpmp as gp


def generate_data():
    """
    Data generation
    (xt, zt): target
    (xi, zi): input dataset
    """
    # Build (xt, zt)
    dim = 1
    nt = 200
    box = [[-1], [1]]
    xt = gp.misc.designs.regulargrid(dim, nt, box)
    zt = gp.misc.testfunctions.twobumps(xt)

    shuffle = True
    if shuffle:
        ni = 5
        ind = np.random.choice(nt, ni, replace=False)
    else:
        ind = [10, 45, 100, 130, 160]
    xi = xt[ind]
    zi = zt[ind]

    return xt, zt, xi, zi


def zero_mean(x, param):
    return None


def constant_mean(x, param):
    return gnp.ones((x.shape[0], 1))


def kernel_ii_or_tt(x, param, pairwise=False):
    """Covariance between observations or predictands at x."""
    p = 2
    sigma2 = gnp.exp(param[0])
    loginvrho = param[1]
    nugget = 100 * gnp.eps

    if pairwise:
        # return a vector of covariances
        K = sigma2 * gnp.ones((x.shape[0], ))  # nx x 0
    else:
        # return a covariance matrix
        K = gnp.scaled_distance(loginvrho, x, x)  # nx x nx
        K = sigma2 * gp.kernel.maternp_kernel(p, K) + nugget * gnp.eye(K.shape[0])

    return K


def kernel_it(x, y, param, pairwise=False):
    """Covariance between observations and prediction points."""
    p = 2
    sigma2 = gnp.exp(param[0])
    loginvrho = param[1]

    if pairwise:
        # return a vector of covariances
        K = gnp.scaled_distance_elementwise(loginvrho, x, y)  # nx x 0
    else:
        # return a covariance matrix
        K = gnp.scaled_distance(loginvrho, x, y)  # nx x ny

    K = sigma2 * gp.kernel.maternp_kernel(p, K)
    return K


def kernel(x, y, param, pairwise=False):
    if y is x or y is None:
        return kernel_ii_or_tt(x, param, pairwise)
    else:
        return kernel_it(x, y, param, pairwise)


def visualize(xt, zt, xi, zi, zpm, zpv):
    fig = gp.misc.plotutils.Figure(isinteractive=True)
    fig.plot(xt, zt, 'C0', linestyle=(0, (5, 5)), linewidth=1.0)
    fig.plotdata(xi, zi)
    fig.plotgp(xt, zpm, zpv)
    fig.xylabels('x', 'z')
    fig.show(grid=True, legend=True, legend_fontsize=9)


def main():
    xt, zt, xi, zi = generate_data()
    mean = constant_mean
    meanparam = None

    covparam = gnp.array([math.log(0.5**2),    # log(sigma2)
                          math.log(1 / .7)])   # log(1/rho)

    model = gp.core.Model(mean, kernel, meanparam, covparam)

    zpm, zpv = model.predict(xi, zi, xt)
    visualize(xt, zt, xi, zi, zpm, zpv)


if __name__ == "__main__":
    main()



# coding: utf-8
import timeit
import jax.numpy as jnp
from jax import grad, jit, vmap
from jax import random

###

key = random.PRNGKey(0)
x = random.normal(key, (10,))
print(x)


###

size = 3000
x = random.normal(key, (size, size), dtype=jnp.float32)
tic = timeit.default_timer();
jnp.dot(x, x.T)  # runs on the CPU
toc = timeit.default_timer()
print(toc-tic)

def sum_logistic(x):
  return jnp.sum(1.0 / (1.0 + jnp.exp(-x)))

x_small = jnp.arange(3.)
derivative_fn = grad(sum_logistic)
print(derivative_fn(x_small))


###
def sigmoid(x):
    return 0.5 * (jnp.tanh(x / 2) + 1)

# Outputs probability of a label being true.
def predict(W, b, inputs):
    return sigmoid(jnp.dot(inputs, W) + b)

# Build a toy dataset.
inputs = jnp.array([[0.52, 1.12,  0.77],
                   [0.88, -1.08, 0.15],
                   [0.52, 0.06, -1.30],
                   [0.74, -2.49, 1.39]])
targets = jnp.array([True, True, False, True])

# Training loss is the negative log-likelihood of the training examples.
def loss(W, b):
    preds = predict(W, b, inputs)
    label_probs = preds * targets + (1 - preds) * (1 - targets)
    return -jnp.sum(jnp.log(label_probs))

# Initialize random model coefficients
key, W_key, b_key = random.split(key, 3)
W = random.normal(W_key, (3,))
b = random.normal(b_key, ())

# Differentiate `loss` with respect to the first positional argument:
W_grad = grad(loss, argnums=0)(W, b)
print('W_grad', W_grad)

# Since argnums=0 is the default, this does the same thing:
W_grad = grad(loss)(W, b)
print('W_grad', W_grad)

# But we can choose different values too, and drop the keyword:
b_grad = grad(loss, 1)(W, b)
print('b_grad', b_grad)

# Including tuple values
W_grad, b_grad = grad(loss, (0, 1))(W, b)
print('W_grad', W_grad)
print('b_grad', b_grad)



import unittest
from examples import (
    gpmp_example01_materncov,
    gpmp_example02_1d_interpolation,
    gpmp_example03_2d,
    gpmp_example04_nd,
    gpmp_example05_1d_custom_kernel,
    gpmp_example06_1d_regression,
    gpmp_example10_sample_paths,
    gpmp_example11_sample_paths_noisy_obs,
    gpmp_example20_1d_interpolation_variation_remap,
    gpmp_example21_1d_interpolation_variation_reml,
    gpmp_example22_1d_interpolation_variation_ml
    )
    

class TestExamples(unittest.TestCase):
    def test_01(self):
        gpmp_example01_materncov.main()

    def test_02(self):
        gpmp_example02_1d_interpolation.main()

    def test_03(self):
        gpmp_example03_2d.main()

    def test_04(self):
        gpmp_example04_nd.main()

    def test_05(self):
        gpmp_example05_1d_custom_kernel.main()

    def test_06(self):
        gpmp_example06_1d_regression.main()

    def test_10(self):
        gpmp_example10_sample_paths.main()

    def test_11(self):
        gpmp_example11_sample_paths_noisy_obs.main()

    def test_20(self):
        gpmp_example20_1d_interpolation_variation_remap.main()
        
    def test_21(self):
        gpmp_example21_1d_interpolation_variation_reml.main()

    def test_22(self):
        gpmp_example22_1d_interpolation_variation_ml.main()


if __name__ == "__main__":
    unittest.main()
